/*
 * gkclient.h
 *
 * Gatekeeper client protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: gkclient.h,v $
 * Revision 1.1  1999/08/31 12:34:18  robertj
 * Added gatekeeper support.
 *
 */

#ifndef __GKCLIENT_H
#define __GKCLIENT_H

#include "transports.h"


class H225_GatekeeperRequest;
class H225_GatekeeperConfirm;
class H225_GatekeeperReject;
class H225_RegistrationRequest;
class H225_RegistrationConfirm;
class H225_RegistrationReject;
class H225_UnregistrationRequest;
class H225_UnregistrationConfirm;
class H225_UnregistrationReject;
class H225_AdmissionRequest;
class H225_AdmissionConfirm;
class H225_AdmissionReject;
class H225_BandwidthRequest;
class H225_BandwidthConfirm;
class H225_BandwidthReject;
class H225_DisengageRequest;
class H225_DisengageConfirm;
class H225_DisengageReject;
class H225_LocationRequest;
class H225_LocationConfirm;
class H225_LocationReject;
class H225_InfoRequest;
class H225_InfoRequestResponse;
class H225_NonStandardMessage;
class H225_UnknownMessageResponse;
class H225_RequestInProgress;
class H225_ResourcesAvailableIndicate;
class H225_ResourcesAvailableConfirm;
class H225_InfoRequestAck;
class H225_InfoRequestNak;



class H323Gatekeeper : public PObject
{
  PCLASSINFO(H323Gatekeeper, PObject);
  public:
  /**@name Construction */
  //@{
    /**Create a new gatekeeper.
     */
    H323Gatekeeper(
      H323EndPoint & endpoint,  /// Endpoint gatekeeper is associated with.
      H323Transport * transport       /// Transport over which gatekeepers communicates.
    );

    /**Destroy gatekeeper.
     */
    ~H323Gatekeeper();

    /**Background thread handler.
      */
    virtual void HandleRasChannel();

    /**Handle and dispatch a RAS PDU..
      */
    virtual BOOL HandleRasPDU(
      const H323RasPDU & response
    );
  //@}

  /**@name Overrides from PObject */
  //@{
    void PrintOn(ostream & strm) const;
  //@}

  /**@name Protocol operations */
  //@{
    /**Discover a gatekeeper on the local network.
     */
    BOOL Discover();

    /**Register with gatekeeper.
       If the address string is empty then it is assumed that the transport
       is already connected to the gatekeeper via the Discover() function.
     */
    BOOL RegistrationRequest(
      const H323TransportAddress & address /// Address of gatekeeper.
    );

    /**Unregister with gatekeeper.
     */
    BOOL UnregistrationRequest(
      int reason      /// Reason for unregistration
    );

    /**Location request to gatekeeper.
     */
    BOOL LocationRequest(
      const PString & alias,          /// Alias name we wish to find.
      H323TransportAddress & address  /// Resultant transport address.
    );

    /**Location request to gatekeeper.
     */
    BOOL LocationRequest(
      const PStringList & aliases,    /// Alias names we wish to find.
      H323TransportAddress & address  /// Resultant transport address.
    );

    /**Admission request to gatekeeper.
     */
    BOOL AdmissionRequest(
      H323Connection & connection  /// Connection we wish to change.
    );


    /**Admission request to gatekeeper.
     */
    BOOL AdmissionRequest(
      H323Connection & connection,    /// Connection we wish to change.
      H323TransportAddress & address  /// Gatekeeper routed transport address.
    );

    /**Disengage request to gatekeeper.
     */
    BOOL DisengageRequest(
      const H323Connection & connection,  /// Connection we wish admitted.
      unsigned reason                     /// Reason code for disengage
    );

    /**Bandwidth request to gatekeeper.
     */
    BOOL BandwidthRequest(
      H323Connection & connection,    /// Connection we wish to change.
      unsigned requestedBandwidth     /// New bandwidth wanted in 0.1kbps
    );

    /**Send an info response to the gatekeeper.
     */
    BOOL InfoRequestResponse(
      const H323Connection * connection,  /// Connection to send info about
      unsigned seqNum = 1
    );
  //@}

  /**@name Protocol callbacks */
  //@{
#define DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(name) \
    virtual void OnSend##name(H225_##name &); \
    virtual BOOL OnReceive##name(const H225_##name &);

    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestResponse);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(NonStandardMessage);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnknownMessageResponse);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RequestInProgress);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(ResourcesAvailableIndicate);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(ResourcesAvailableConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestAck);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestNak);

#undef DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS

    /**Handle unknown PDU type.
      */
    virtual void OnReceiveUnkown(
      const H323RasPDU & pdu  /// PDU that was not handled.
    );
  //@}

  /**@name Member variable access */
  //@{
    /**Determine if the endpoint is registered with the gatekeeper.
      */
    BOOL IsRegistered() const { return isRegistered; }
  //@}


  protected:
    BOOL MakeRequest(const H323RasPDU & request);
    BOOL CheckForResponse(unsigned, unsigned);


    H323EndPoint  & endpoint;
    H323Transport * transport;

    PMutex     makeRequestMutex;
    unsigned   lastSequenceNumber;
    unsigned   requestTag;
    PSyncPoint responseHandled;
    enum {
      ConfirmReceived,
      RejectReceived,
      RequestInProgress
    } responseResult;

    BOOL    isRegistered;
    PString gatekeeperIdentifier;
    PString endpointIdentifier;
    PTimer  timeToLive;

    unsigned             allocatedBandwidth;
    H323TransportAddress locatedAddress;
};


PLIST(H323GatekeeperList, H323Gatekeeper);


#endif // __GKCLIENT_H


/////////////////////////////////////////////////////////////////////////////
