/*
 * rtp.h
 *
 * RTP protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: rtp.h,v $
 * Revision 1.5  1999/08/31 12:34:18  robertj
 * Added gatekeeper support.
 *
 * Revision 1.4  1999/07/13 09:53:24  robertj
 * Fixed some problems with jitter buffer and added more debugging.
 *
 * Revision 1.3  1999/07/09 06:09:49  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.2  1999/06/22 13:49:40  robertj
 * Added GSM support and further RTP protocol enhancements.
 *
 * Revision 1.1  1999/06/14 06:12:25  robertj
 * Changes for using RTP sessions correctly in H323 Logical Channel context
 *
 */

#ifndef __RTP_H
#define __RTP_H


#include <ptlib/sockets.h>


class H225_RTPSession;

class H245_H2250LogicalChannelParameters;
class H245_H2250LogicalChannelAckParameters;

class H323Connection;


///////////////////////////////////////////////////////////////////////////////
// Real Time Protocol - IETF RFC1889 and RFC1890

class RTP_DataFrame : public PBYTEArray
{
  PCLASSINFO(RTP_DataFrame, PBYTEArray);

  public:
    RTP_DataFrame(PINDEX payloadSize = 2048);

    enum {
      ProtocolVersion = 2,
      MinHeaderSize = 12
    };

    enum PayloadTypes {
      PCMU,
      FS1016,
      G721,
      GSM,
      G7231,
      DVI4_8k,
      DVI4_16k,
      LPC,
      PCMA,
      G722,
      L16_Stereo,
      L16_Mono,
      MPA = 14,
      G728,
      CelB = 25,
      JPEG,
      H261 = 31,
      MPV,
      MP2T,
      DynamicBase = 96,
      MaxPayloadType = 127
    };

    unsigned GetVersion() const { return (theArray[0]>>6)&3; }

    BOOL GetExtension() const   { return (theArray[0]&0x10) != 0; }
    void SetExtension(BOOL ext);

    BOOL GetMarker() const { return (theArray[1]&0x80) != 0; }
    void SetMarker(BOOL m);

    unsigned GetPayloadType() const { return  theArray[1]&0x7f; }
    void     SetPayloadType(unsigned t);

    WORD GetSequenceNumber() const { return *(PUInt16b *)&theArray[2]; }
    void SetSequenceNumber(WORD n) { *(PUInt16b *)&theArray[2] = n; }

    DWORD GetTimestamp() const  { return *(PUInt32b *)&theArray[4]; }
    void  SetTimestamp(DWORD t) { *(PUInt32b *)&theArray[4] = t; }

    DWORD GetSyncSource() const  { return *(PUInt32b *)&theArray[8]; }
    void  SetSyncSource(DWORD s) { *(PUInt32b *)&theArray[8] = s; }

    PINDEX GetContribSrcCount() const { return theArray[0]&0xf; }
    DWORD  GetContribSource(PINDEX idx) const;
    void   SetContribSource(PINDEX idx, DWORD src);

    PINDEX GetHeaderSize()     const { return 12 + 4*GetContribSrcCount(); }

    PINDEX GetPayloadSize() const { return payloadSize; }
    BOOL   SetPayloadSize(PINDEX sz);
    BYTE * GetPayloadPtr()     const { return (BYTE *)(theArray+GetHeaderSize()); }

  protected:
    PINDEX payloadSize;
};


class RTP_ControlFrame : public PBYTEArray
{
  PCLASSINFO(RTP_ControlFrame, PBYTEArray);

  public:
    RTP_ControlFrame(PINDEX payloadSize = 2048);

    unsigned GetVersion() const { return (BYTE)theArray[0]>>6; }
    unsigned GetCount()   const { return (BYTE)theArray[0]&0x1f; }

    enum PayloadTypes {
      e_SenderReport = 200,
      e_ReceiverReport,
      e_SourceDescription,
      e_Goodbye,
      e_ApplDefined
    };

    unsigned GetPayloadType() const { return (BYTE)theArray[1]; }
    void     SetPayloadType(unsigned t);

    PINDEX GetPayloadSize() const { return 4*(*(PUInt16b *)&theArray[2]); }
    void   SetPayloadSize(PINDEX sz);

    BYTE * GetPayloadPtr() const { return (BYTE *)(theArray+4); }

#pragma pack(1)
    struct ReceptionReport {
      PUInt32b ssrc;      /* data source being reported */
      BYTE fraction;      /* fraction lost since last SR/RR */
      BYTE lost[3];	  /* cumulative number of packets lost (signed!) */
      PUInt32b last_seq;  /* extended last sequence number received */
      PUInt32b jitter;    /* interarrival jitter */
      PUInt32b lsr;       /* last SR packet from this source */
      PUInt32b dlsr;      /* delay since last SR packet */

      unsigned GetLostPackets() const { return (lost[0]<<16U)+(lost[1]<<8U)+lost[2]; }
    };

    struct SenderReport {
      PUInt32b ssrc;      /* source this RTCP packet refers to */
      PUInt32b ntp_sec;   /* NTP timestamp */
      PUInt32b ntp_frac;
      PUInt32b rtp_ts;    /* RTP timestamp */
      PUInt32b psent;     /* packets sent */
      PUInt32b osent;     /* octets sent */ 
    };

    struct SourceDescription {
      PUInt32b src;       /* first SSRC/CSRC */
      struct Item {
        BYTE type;        /* type of SDES item (rtcp_sdes_type_t) */
        BYTE length;      /* length of SDES item (in octets) */
        char data[1];     /* text, not zero-terminated */
      } item[1];          /* list of SDES items */
    };
#pragma pack()
};


/**This class is for encpsulating the IETF Real Time Protocol interface.
 */
class RTP_Session : public PObject
{
  PCLASSINFO(RTP_Session, PObject);

  public:
    enum {
      DefaultAudioSessionID = 1,
      DefaultVideoSessionID = 2
    };

  /**@name Construction */
  //@{
    /**Create a new channel.
     */
    RTP_Session(
      const H323Connection & connection, /// Owner of the RTP session
      unsigned id                        /// Session ID for RTP channel
    );
  //@}

  /**@name Operations */
  //@{
    /**Read a data frame from the RTP channel.
       Any control frames received are dispatched to callbacks and are not
       returned by this function. It will block until a data frame is
       available or an error occurs.
      */
    virtual BOOL ReadData(RTP_DataFrame & frame) = 0;

    /**Write a data frame from the RTP channel.
      */
    virtual BOOL WriteData(RTP_DataFrame & frame) = 0;

    /**Write a control frame from the RTP channel.
      */
    virtual BOOL WriteControl(RTP_ControlFrame & frame) = 0;

    /**Close down the RTP session.
      */
    virtual void Close() = 0;

    /**Fill out the OpenLogicalChannel PDU for the particular channel type.
     */
    virtual BOOL OnSendingPDU(
      H245_H2250LogicalChannelParameters & param, /// Open PDU to send.
      BOOL includeMediaChannel                    /// Include media channel field
    ) const = 0;

    /**This is called when request to create a channel is received from a
       remote machine and is about to be acknowledged.
     */
    virtual void OnSendingAckPDU(
      H245_H2250LogicalChannelAckParameters & param  /// Acknowledgement PDU
    ) const = 0;

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.
     */
    virtual BOOL OnReceivedPDU(
      const H245_H2250LogicalChannelParameters & param, /// Acknowledgement PDU
      unsigned & errorCode                              /// Error on failure
    ) = 0;

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.
     */
    virtual BOOL OnReceivedAckPDU(
      const H245_H2250LogicalChannelAckParameters & param  /// Acknowledgement PDU
    ) = 0;

    /**This is called when a gatekeeper wants to get status information from
       the endpoint.

       The default behaviour fills in the session ID's and SSRC parameters
       but does not do anything with the transport fields.
     */
    virtual void OnSendRasInfo(
      H225_RTPSession & info  /// RTP session info PDU
    );
  //@}

  /**@name Call back frunctions */
  //@{
    virtual BOOL OnSendData(RTP_DataFrame & frame);
    virtual BOOL OnReceiveData(const RTP_DataFrame & frame);
    virtual BOOL OnReceiveControl(const RTP_ControlFrame & frame);

    virtual void OnRxSenderReport(const RTP_ControlFrame::SenderReport & sender,
                                  const RTP_ControlFrame::ReceptionReport * reports,
                                  PINDEX count);
    virtual void OnRxReceiverReport(DWORD src,
                                    const RTP_ControlFrame::ReceptionReport * reports,
                                    PINDEX count);
    virtual void OnRxSourceDescription(const RTP_ControlFrame::SourceDescription & description);
    virtual void OnRxGoodbye(const PUInt32b * ssrc, PINDEX count,
                             const PString & reason);
    virtual void OnRxApplDefined(const PString & type, unsigned subtype, DWORD src,
                                 const BYTE * data, PINDEX size);
  //@}

  /**@name Member variable access */
  //@{
    /**Get the ID for the RTP session.
      */
    unsigned GetSessionID() const { return sessionID; }

    /**Increment reference count for RTP session.
      */
    void IncrementReference() { referenceCount++; }

    /**Decrement reference count for RTP session.
      */
    BOOL DecrementReference() { return --referenceCount == 0; }
  //@}

  protected:
    const H323Connection & connection;

    BYTE     sessionID;
    unsigned referenceCount;

    DWORD    syncSourceOut;
    unsigned senderReportInterval;

    WORD     lastSequenceNumber;
    WORD     expectedSequenceNumber;

    DWORD    packetsSent;
    DWORD    octetsSent;

    DWORD    packetsReceived;
    DWORD    packetsLost;
    DWORD    jitterLevel;
};

PDICTIONARY(RTPSessionDict, POrdinalKey, RTP_Session);


/**This class is for the IETF Real Time Protocol interface on UDP/IP.
 */
class RTP_UDP : public RTP_Session
{
  PCLASSINFO(RTP_UDP, RTP_Session);

  public:
  /**@name Construction */
  //@{
    /**Create a new RTP channel.
     */
    RTP_UDP(
      const H323Connection & connection, /// Owner of the RTP session
      unsigned id,  /// Session ID for RTP channel
      WORD portBase /// Base port number for socket pair
    );

    /// Destroy the RTP
    ~RTP_UDP();
  //@}

  /**@name Overrides from class RTP_Session */
  //@{
    /**Read a data frame from the RTP channel.
       Any control frames received are dispatched to callbacks and are not
       returned by this function. It will block until a data frame is
       available or an error occurs.
      */
    virtual BOOL ReadData(RTP_DataFrame & frame);

    /**Write a data frame from the RTP channel.
      */
    virtual BOOL WriteData(RTP_DataFrame & frame);

    /**Write a control frame from the RTP channel.
      */
    virtual BOOL WriteControl(RTP_ControlFrame & frame);

    /**Close down the RTP session.
      */
    virtual void Close();

    /**Fill out the OpenLogicalChannel PDU for the particular channel type.
     */
    virtual BOOL OnSendingPDU(
      H245_H2250LogicalChannelParameters & param, /// Open PDU to send.
      BOOL includeMediaChannel                    /// Include media channel field
    ) const;

    /**This is called when request to create a channel is received from a
       remote machine and is about to be acknowledged.
     */
    virtual void OnSendingAckPDU(
      H245_H2250LogicalChannelAckParameters & param  /// Acknowledgement PDU
    ) const;

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.

       The default behaviour sets the remote ports to send UDP packets to.
     */
    virtual BOOL OnReceivedPDU(
      const H245_H2250LogicalChannelParameters & param, /// Acknowledgement PDU
      unsigned & errorCode                              /// Error on failure
    );

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.

       The default behaviour sets the remote ports to send UDP packets to.
     */
    virtual BOOL OnReceivedAckPDU(
      const H245_H2250LogicalChannelAckParameters & param  /// Acknowledgement PDU
    );

    /**This is called when a gatekeeper wants to get status information from
       the endpoint.

       The default behaviour calls the ancestor functon and then fills in the 
       transport fields.
     */
    virtual void OnSendRasInfo(
      H225_RTPSession & info  /// RTP session info PDU
    );
  //@}

  protected:
    BOOL ReadDataPDU(RTP_DataFrame & frame);
    BOOL ReadControlPDU();

    PIPSocket::Address remoteAddress;
    WORD               remoteDataPort;
    WORD               remoteControlPort;

    PUDPSocket dataSocket;
    PUDPSocket controlSocket;
};


#endif // __RTP_H


/////////////////////////////////////////////////////////////////////////////
