/*
 * channels.cxx
 *
 * H.323 protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: channels.cxx,v $
 * Revision 1.23  1999/09/03 02:17:50  robertj
 * Added more debugging
 *
 * Revision 1.22  1999/08/31 12:34:18  robertj
 * Added gatekeeper support.
 *
 * Revision 1.21  1999/07/16 15:02:51  robertj
 * Changed jitter buffer to throw away old packets if jitter exceeded.
 *
 * Revision 1.20  1999/07/16 00:51:03  robertj
 * Some more debugging of fast start.
 *
 * Revision 1.19  1999/07/15 14:45:35  robertj
 * Added propagation of codec open error to shut down logical channel.
 * Fixed control channel start up bug introduced with tunnelling.
 *
 * Revision 1.18  1999/07/15 09:04:31  robertj
 * Fixed some fast start bugs
 *
 * Revision 1.17  1999/07/14 06:04:04  robertj
 * Fixed setting of channel number in fast start.
 *
 * Revision 1.16  1999/07/13 09:53:24  robertj
 * Fixed some problems with jitter buffer and added more debugging.
 *
 * Revision 1.15  1999/07/13 02:50:58  craigs
 * Changed semantics of SetPlayDevice/SetRecordDevice, only descendent
 *    endpoint assumes PSoundChannel devices for audio codec.
 *
 * Revision 1.14  1999/07/10 03:01:48  robertj
 * Removed debugging.
 *
 * Revision 1.13  1999/07/09 06:09:49  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.12  1999/06/25 14:19:40  robertj
 * Fixed termination race condition in logical channel tear down.
 *
 * Revision 1.11  1999/06/24 13:32:45  robertj
 * Fixed ability to change sound device on codec and fixed NM3 G.711 compatibility
 *
 * Revision 1.10  1999/06/22 13:49:40  robertj
 * Added GSM support and further RTP protocol enhancements.
 *
 * Revision 1.9  1999/06/14 05:15:55  robertj
 * Changes for using RTP sessions correctly in H323 Logical Channel context
 *
 * Revision 1.8  1999/06/13 12:41:14  robertj
 * Implement logical channel transmitter.
 * Fixed H245 connect on receiving call.
 *
 * Revision 1.7  1999/06/09 06:18:00  robertj
 * GCC compatibiltiy.
 *
 * Revision 1.6  1999/06/09 05:26:19  robertj
 * Major restructuring of classes.
 *
 * Revision 1.5  1999/06/07 00:54:30  robertj
 * Displayed error on SetOption for buffer size
 *
 * Revision 1.4  1999/06/06 06:06:36  robertj
 * Changes for new ASN compiler and v2 protocol ASN files.
 *
 * Revision 1.3  1999/04/26 06:14:46  craigs
 * Initial implementation for RTP decoding and lots of stuff
 * As a whole, these changes are called "First Noise"
 *
 * Revision 1.2  1999/02/25 03:26:02  robertj
 * BeOS compatibility
 *
 * Revision 1.1  1999/01/16 01:31:09  robertj
 * Initial revision
 *
 */

#include <ptlib.h>
#include "channels.h"
#include "h323pdu.h"
#include "h323.h"


//#define JITTER_DEBUG 10000


class H323LogicalChannelThread : public PThread
{
    PCLASSINFO(H323LogicalChannelThread, PThread)
  public:
    H323LogicalChannelThread(H323EndPoint & endpoint, H323Channel & channel, BOOL rx);
    void Main();
  private:
    H323Channel & channel;
    BOOL receiver;
};


class RTP_JitterBuffer : public PThread
{
  PCLASSINFO(RTP_JitterBuffer, PThread);

  public:
    RTP_JitterBuffer(
      H323EndPoint & endpoint,
      H323Codec & codec
    );
    ~RTP_JitterBuffer();

    virtual void Main();

    RTP_DataFrame * GetReadFrame();
    void QueueReadFrame();
    RTP_DataFrame * GetWriteFrame(DWORD timestamp);
    void FreeWriteFrame();

    PINDEX GetSize() const { return bufferSize; }

    void PreBuffering() { preBuffering = TRUE; }

  protected:
    class Entry : public RTP_DataFrame
    {
      public:
        Entry(PINDEX size) { SetPayloadSize(size); }
        Entry * next;
        Entry * prev;
    };

    H323Codec & codec;
    PINDEX      bufferSize;
    DWORD       maxJitterTime;
    Entry     * oldestFrame;
    Entry     * newestFrame;
    Entry     * freeFrames;
    Entry     * currentReadFrame;
    Entry     * currentWriteFrame;
    PMutex      mutex;
    BOOL        running;
    BOOL        preBuffering;

#ifdef JITTER_DEBUG
    PTimeInterval debugInTick[JITTER_DEBUG], debugOutTick[JITTER_DEBUG];
    DWORD debugInTime[JITTER_DEBUG], debugOutTime[JITTER_DEBUG];
    int debugInPos, debugOutPos;
#endif
};


#define new PNEW


/////////////////////////////////////////////////////////////////////////////

H323LogicalChannelThread::H323LogicalChannelThread(H323EndPoint & endpoint,
                                                   H323Channel & c,
                                                   BOOL rx)
  : PThread(endpoint.GetChannelThreadStackSize(), NoAutoDeleteThread),
    channel(c)
{
  receiver = rx;
  Resume();
}


void H323LogicalChannelThread::Main()
{
  if (receiver)
    channel.Receive();
  else
    channel.Transmit();
}


/////////////////////////////////////////////////////////////////////////////

H323ChannelNumber::H323ChannelNumber(unsigned num, BOOL fromRem)
{
  PAssert(num < 0x10000, PInvalidParameter);
  number = num;
  fromRemote = fromRem;
}


PObject * H323ChannelNumber::Clone() const
{
  return new H323ChannelNumber(number, fromRemote);
}


PINDEX H323ChannelNumber::HashFunction() const
{
  if (fromRemote)
    return number;

  return number|0x10000;
}


void H323ChannelNumber::PrintOn(ostream & strm) const
{
  strm << (fromRemote ? 'R' : 'T') << '-' << number;
}


H323ChannelNumber & H323ChannelNumber::operator++(int)
{
  number++;
  return *this;
}


/////////////////////////////////////////////////////////////////////////////

H323Channel::H323Channel(H323Connection & conn, const H323Capability & cap)
  : endpoint(conn.GetEndPoint()),
    connection(conn),
    capability(cap)
{
  codec = NULL;
  bandwidthUsed = 0;
  receiveThread = NULL;
  transmitThread = NULL;
}


H323Channel::~H323Channel()
{
  connection.UseBandwidth(bandwidthUsed, FALSE);

  if (codec != NULL)
    codec->Close();

  if (receiveThread != NULL)
    receiveThread->WaitForTermination();

  if (transmitThread != NULL)
    transmitThread->WaitForTermination();

  delete codec;
  delete receiveThread;
  delete transmitThread;
}


unsigned H323Channel::GetSessionID() const
{
  return 0;
}


BOOL H323Channel::IsRunning() const
{
  if (receiveThread  != NULL && !receiveThread ->IsTerminated())
    return TRUE;

  if (transmitThread != NULL && !transmitThread->IsTerminated())
    return TRUE;

  return FALSE;
}


BOOL H323Channel::OnReceivedPDU(const H245_OpenLogicalChannel & /*pdu*/,
                                unsigned & /*errorCode*/)
{
  return TRUE;
}


BOOL H323Channel::OnReceivedAckPDU(const H245_OpenLogicalChannelAck & /*pdu*/)
{
  return TRUE;
}


void H323Channel::OnSendOpenAck(const H245_OpenLogicalChannel & /*pdu*/,
                                H245_OpenLogicalChannelAck & /* pdu*/) const
{
}


void H323Channel::OnFlowControl(long bitRateRestriction)
{
  PTRACE(3, "LogCh\tOnFlowControl: " << bitRateRestriction);
}


void H323Channel::OnMiscellaneousCommand(const H245_MiscellaneousCommand_type & type)
{
  PTRACE(3, "LogCh\tOnMiscellaneousCommand: " << type.GetTagName());
}


void H323Channel::OnMiscellaneousIndication(const H245_MiscellaneousIndication_type & type)
{
  PTRACE(3, "LogCh\tOnMiscellaneousIndication: " << type.GetTagName());
}


void H323Channel::OnJitterIndication(DWORD jitter,
                                     int skippedFrameCount,
                                     int additionalBuffer)
{
  PTRACE(3, "LogCh\tOnJitterIndication:"
            " jitter=" << jitter <<
            " skippedFrameCount=" << skippedFrameCount <<
            " additionalBuffer=" << additionalBuffer);
}


BOOL H323Channel::SetInitialBandwidth()
{
  if (codec != NULL)
    return SetBandwidthUsed(codec->GetBandwidth());

  return FALSE;
}


BOOL H323Channel::SetBandwidthUsed(unsigned bandwidth)
{
  connection.UseBandwidth(bandwidthUsed, FALSE);
  bandwidthUsed = 0;

  if (!connection.UseBandwidth(bandwidth, TRUE))
    return FALSE;

  bandwidthUsed = bandwidth;
  return TRUE;
}


/////////////////////////////////////////////////////////////////////////////

H323UnidirectionalChannel::H323UnidirectionalChannel(H323Connection & conn,
                                                     const H323Capability & cap,
                                                     Directions direction)
  : H323Channel(conn, cap),
    receiver(direction == IsReceiver)
{
  codec = capability.CreateCodec(receiver ? H323Codec::Decoder : H323Codec::Encoder);
}


H323Channel::Directions H323UnidirectionalChannel::GetDirection() const
{
  return receiver ? IsReceiver : IsTransmitter;
}


void H323UnidirectionalChannel::Start()
{
  PThread * thread = new H323LogicalChannelThread(endpoint, *this, receiver);

  if (receiver)
    receiveThread  = thread;
  else
    transmitThread = thread;
}


/////////////////////////////////////////////////////////////////////////////

H323BidirectionalChannel::H323BidirectionalChannel(H323Connection & conn,
                                                   const H323Capability & cap)
  : H323Channel(conn, cap)
{
}


H323Channel::Directions H323BidirectionalChannel::GetDirection() const
{
  return IsBidirectional;
}


void H323BidirectionalChannel::Start()
{
  receiveThread  = new H323LogicalChannelThread(endpoint, *this, TRUE);
  transmitThread = new H323LogicalChannelThread(endpoint, *this, FALSE);
}


/////////////////////////////////////////////////////////////////////////////

H323_RTPChannel::H323_RTPChannel(H323Connection & conn,
                                 const H323Capability & cap,
                                 Directions direction,
                                 RTP_Session & r)
  : H323UnidirectionalChannel(conn, cap, direction),
    rtp(r)
{
  PTRACE(3, "H323_RTP\t" << (receiver ? "Receiver" : "Transmitter")
         << " created using session " << rtp.GetSessionID());
}


H323_RTPChannel::~H323_RTPChannel()
{
  if (receiveThread != NULL) {
    // Break any I/O blocks and wait for the thread that uses this object to
    // terminate before we allow it to be deleted.
    rtp.Close();
    receiveThread->WaitForTermination();
  }

  if (transmitThread != NULL) {
    codec->Close();
    transmitThread->WaitForTermination();
  }

  // Finished with the RTP session, this will delete the session if it is no
  // longer referenced by any logical channels.
  connection.ReleaseSession(rtp.GetSessionID());
}


unsigned H323_RTPChannel::GetSessionID() const
{
  return rtp.GetSessionID();
}


BOOL H323_RTPChannel::OnSendingPDU(H245_OpenLogicalChannel & open) const
{
  PTRACE(3, "H323_RTP\tOnSendingPDU");

  open.m_forwardLogicalChannelNumber = (unsigned)number;

  if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
    open.m_reverseLogicalChannelParameters.IncludeOptionalField(
            H245_OpenLogicalChannel_reverseLogicalChannelParameters::e_multiplexParameters);
    // Set the communications information for unicast IPv4
    open.m_reverseLogicalChannelParameters.m_multiplexParameters.SetTag(
                H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters
                    ::e_h2250LogicalChannelParameters);

    return rtp.OnSendingPDU(open.m_reverseLogicalChannelParameters.m_multiplexParameters,
                            receiver);
  }
  else {
    // Set the communications information for unicast IPv4
    open.m_forwardLogicalChannelParameters.m_multiplexParameters.SetTag(
                H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters
                    ::e_h2250LogicalChannelParameters);

    return rtp.OnSendingPDU(open.m_forwardLogicalChannelParameters.m_multiplexParameters,
                            receiver);
  }
}


void H323_RTPChannel::OnSendOpenAck(const H245_OpenLogicalChannel & open,
                                    H245_OpenLogicalChannelAck & ack) const
{
  PTRACE(3, "H323_RTP\tOnSendOpenAck");

  // set forwardMultiplexAckParameters option
  ack.IncludeOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters);

  // select H225 choice
  ack.m_forwardMultiplexAckParameters.SetTag(
	  H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters);

  // get H225 parms
  H245_H2250LogicalChannelAckParameters & param = ack.m_forwardMultiplexAckParameters;

  // set session ID
  param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_sessionID);
  const H245_H2250LogicalChannelParameters & openparam =
                          open.m_forwardLogicalChannelParameters.m_multiplexParameters;
  unsigned sessionID = openparam.m_sessionID;
  param.m_sessionID = sessionID;

  rtp.OnSendingAckPDU(param);

  PTRACE(2, "H323_RTP\tSending open logical channel ACK: sessionID=" << sessionID);
}


BOOL H323_RTPChannel::OnReceivedPDU(const H245_OpenLogicalChannel & open,
                                    unsigned & errorCode)
{
  if (receiver)
    number = H323ChannelNumber(open.m_forwardLogicalChannelNumber, TRUE);

  PTRACE(3, "H323_RTP\tOnReceivedPDU for channel: " << number);

  if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
    if (open.m_reverseLogicalChannelParameters.m_multiplexParameters.GetTag() ==
             H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters::
                                     e_h2250LogicalChannelParameters)
      return rtp.OnReceivedPDU(open.m_reverseLogicalChannelParameters.m_multiplexParameters, errorCode);
  }
  else {
    if (open.m_forwardLogicalChannelParameters.m_multiplexParameters.GetTag() ==
             H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters::
                                     e_h2250LogicalChannelParameters)
      return rtp.OnReceivedPDU(open.m_forwardLogicalChannelParameters.m_multiplexParameters, errorCode);
  }

  PTRACE(1, "H323_RTP\tOnly H.225.0 multiplex supported");
  errorCode = H245_OpenLogicalChannelReject_cause::e_unsuitableReverseParameters;
  return FALSE;
}


BOOL H323_RTPChannel::OnReceivedAckPDU(const H245_OpenLogicalChannelAck & ack)
{
  PTRACE(3, "H323_RTP\tOnReceiveOpenAck");

  if (!ack.HasOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters)) {
    PTRACE(1, "H323_RTP\tNo forwardMultiplexAckParameters");
    return FALSE;
  }

  if (ack.m_forwardMultiplexAckParameters.GetTag() !=
            H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters) {
    PTRACE(1, "H323_RTP\tOnly H.225.0 multiplex supported");
    return FALSE;
  }

  return rtp.OnReceivedAckPDU(ack.m_forwardMultiplexAckParameters);
}


void H323_RTPChannel::Transmit()
{
  if (codec == NULL) {
    PAssertAlways(PLogicError);
    return;
  }

  // Give the connection (or endpoint) a chance to do something with
  // the opening of the codec. Default calls codec->Open();
  if (!connection.OnStartLogicalChannel(*this))
    return;

  // Get parameters from the codec on time and data sizes
  RTP_DataFrame frame(codec->GetMaxFrameSize());
  unsigned codecTimestamps = codec->GetFrameRate();
  unsigned codecMilliseconds = codecTimestamps/codec->GetTimeUnits();

  PTRACE(2, "H323_RTP\tTransmit thread started:"
            " rate=" << codecTimestamps <<
            " time=" << codecMilliseconds << "ms");

  // This is real time so need to keep track of elapsed milliseconds
  PTimeInterval nextTick = PTimer::Tick();
  DWORD timestamp = 0;
  BOOL silent = TRUE;
  unsigned length;

  // Now keep getting encoded frames from the codec, we make sure we do not
  // call it any faster than codecMilliseconds, even though as a rule
  // the function takes exactly codecMilliseconds to complete anyway.

  while (codec->Read(frame.GetPayloadPtr(), length)) {
    // Check for silence from codec.
    if (length == 0)
      silent = TRUE;
    else {
      frame.SetMarker(silent);  // Set flag for start of sound
      frame.SetPayloadSize(length);
      frame.SetTimestamp(timestamp);

      // This allows for the fact that a codec may change the RTP payload
      // type on the fly. If it can be shown that this does not occur then
      // this can be moved to outside of the loop.
      frame.SetPayloadType(codec->GetRTPPayloadType());

      // Send the frame of coded data to RTP transport
      if (!rtp.WriteData(frame))
        break;

      // Is no longer silent
      silent = FALSE;
    }

    // Calculate the timestamp and real time to take in processing
    timestamp += codecTimestamps;
    nextTick += codecMilliseconds;

    // Wait for milliseconds to go past
    PTimeInterval delay = nextTick - PTimer::Tick();
    if (delay > 1)
      PThread::Current()->Sleep(delay);
  }

  PTRACE(2, "H323_RTP\tTransmit thread ended");
}


void H323_RTPChannel::Receive()
{
  PTRACE(2, "H323_RTP\tReceive thread started");

  if (codec == NULL) {
    PAssertAlways(PLogicError);
    return;
  }

  // Give the connection (or endpoint) a chance to do something with
  // the opening of the codec. Default calls codec->Open();
  if (!connection.OnStartLogicalChannel(*this))
    return;

  // Start the thread that is on the other end of the jitter buffer
  RTP_JitterBuffer * jitter = new RTP_JitterBuffer(endpoint, *codec);

  // Get the first free frame in the jitter buffer
  RTP_DataFrame * frame = jitter->GetReadFrame();

  // Keep reading from the RTP transport frames
  while (rtp.ReadData(*frame)) {
    // Make sure is from the current payload type for the codec selected
    if (frame->GetPayloadType() != codec->GetRTPPayloadType())
      PTRACE(2, "H323_RTP\tPayload type mismatch: "
             << frame->GetPayloadType() << " != " << codec->GetRTPPayloadType());
    else {
      if (frame->GetMarker()) {
        PTRACE(3, "H323_RTP\tStart of talk burst: " << frame->GetTimestamp());
        jitter->PreBuffering();
      }

      // Queue the frame for playing by the thread at other end of jitter buffer
      jitter->QueueReadFrame();

      // Get next free frame out of the jitter buffer
      frame = jitter->GetReadFrame();
    }
  }

  delete jitter;

  PTRACE(2, "H323_RTP\tReceive thread ended");
}


/////////////////////////////////////////////////////////////////////////////

RTP_JitterBuffer::RTP_JitterBuffer(H323EndPoint & endpoint, H323Codec & c)
  : PThread(endpoint.GetChannelThreadStackSize(),
            NoAutoDeleteThread,
            HighPriority),
    codec(c)
{
  // Jitter buffer is a queue of frames waiting for playback, a list of
  // free frames, and a couple of place holders for the frame that is
  // currently beeing read from the RTP transport or written to the codec.

  oldestFrame = newestFrame = currentReadFrame = currentWriteFrame = NULL;

  // Calculate the maximum amount of timestamp units for the jitter buffer
  maxJitterTime = endpoint.GetMaxAudioDelayJitter()*codec.GetTimeUnits();

  // Calculate number of frames to allocate
  bufferSize = (maxJitterTime/codec.GetFrameRate() + 1)*2;

  // Allocate the frames and put them all into the free list
  freeFrames = new Entry(codec.GetMaxFrameSize());
  freeFrames->next = freeFrames->prev = NULL;

  for (PINDEX i = 0; i < bufferSize; i++) {
    Entry * frame = new Entry(codec.GetMaxFrameSize());
    frame->prev = NULL;
    frame->next = freeFrames;
    freeFrames->prev = frame;
    freeFrames = frame;
  }

  PTRACE(2, "H323_RTP\tJitter buffer created:"
            " size=" << bufferSize <<
            " timestamp=" << maxJitterTime);

#ifdef JITTER_DEBUG
  debugInPos = debugOutPos = 0;
#endif

  // Start reading data from jitter buffer and writing to codec
  running = TRUE;
  Resume();
}


#ifdef JITTER_DEBUG
static void JitterDebugDump(const char * title, int count, PTimeInterval * tick, DWORD * time)
{
  PTimeInterval prevtick = tick[0];
  DWORD prevtime = time[0];
  int i;

  for (i = 0; i < count; i++) {
    PTRACE(3, title
          << ' ' << tick[i] << ' ' << (tick[i]-prevtick)
          << ' ' << time[i] << ' ' << (time[i]-prevtime));
    prevtick = tick[i];
    prevtime = time[i];
  }
}
#endif


RTP_JitterBuffer::~RTP_JitterBuffer()
{
  running = FALSE;
  WaitForTermination();

  mutex.Wait();

  // Free up all the memory allocated
  while (oldestFrame != NULL) {
    Entry * frame = oldestFrame;
    oldestFrame = oldestFrame->next;
    delete frame;
  }

  while (freeFrames != NULL) {
    Entry * frame = freeFrames;
    freeFrames = freeFrames->next;
    delete frame;
  }

  delete currentReadFrame;
  delete currentWriteFrame;

  mutex.Signal();

#ifdef JITTER_DEBUG
  JitterDebugDump("In ", debugInPos, debugInTick, debugInTime);
  JitterDebugDump("Out", debugOutPos, debugOutTick, debugOutTime);
#endif
}


void RTP_JitterBuffer::Main()
{
  // Get the real time parameters from the codec.
  unsigned codecTimestamps = codec.GetFrameRate();
  unsigned codecMilliseconds = codecTimestamps/codec.GetTimeUnits();

  PTRACE(2, "H323_RTP\tJitter thread started:"
            " rate=" << codecTimestamps <<
            " time=" << codecMilliseconds << "ms");

  // This is real time so need to keep track of elapsed milliseconds
  PTimeInterval nextTick = PTimer::Tick();
  DWORD timestamp = 0;
  RTP_DataFrame * frame = NULL;
  unsigned byteOffset = 0;

  while (running) {
    // Get the next frame from buffer, if it is time to do so and have
    // finished the previous frame.
    if (frame == NULL) {
      frame = GetWriteFrame(timestamp);
      // If have a frame set timestamp to correct value (allows for lost data)
      if (frame != NULL)
        timestamp = frame->GetTimestamp();
    }

    // Not time yet, just wait
    if (frame == NULL) {
      // Output silence
      unsigned dummy;
      if (!codec.Write(NULL, 0, dummy))
        break;
    }
    else {
      // Output noise
      unsigned size = frame->GetPayloadSize();
      unsigned written;
      if (!codec.Write(frame->GetPayloadPtr()+byteOffset, size-byteOffset, written))
        break;

      // Move offset to next (possible) codec data block
      byteOffset += written;

      // Return the frame to the free list if got all of the data out of it
      if (byteOffset >= size) {
        frame = NULL;
        byteOffset = 0;
        FreeWriteFrame();
      }
    }

    // Calculate the timestamp value for next frame
    timestamp += codecTimestamps;

    // Calculate the real time tick we should wait for it to arrive
    nextTick += codecMilliseconds;
    PTimeInterval delay = nextTick - PTimer::Tick();
    if (delay > 0) {
      PAssert(delay < 1000, "Abnormally large jitter time.");
      Sleep(delay);
    }
  }

  PTRACE(2, "H323_RTP\tJitter thread ended");
}


RTP_DataFrame * RTP_JitterBuffer::GetReadFrame()
{
  // Get the next free frame available for use for reading from the RTP
  // transport. Place it into a parking spot.

  if (currentReadFrame != NULL)
    return currentReadFrame;

  mutex.Wait();

  if (freeFrames == NULL) {
    // We have a full jitter buffer, need a new frame so take the oldest one
    currentReadFrame = oldestFrame;
    oldestFrame = oldestFrame->next;
    if (oldestFrame != NULL)
      oldestFrame->prev = NULL;
  }
  else {
    // Take the next free frame and make it the current for reading
    currentReadFrame = freeFrames;
    freeFrames = freeFrames->next;
    if (freeFrames != NULL)
      freeFrames->prev = NULL;
  }

  currentReadFrame->next = NULL;

  mutex.Signal();

  return currentReadFrame;
}


void RTP_JitterBuffer::QueueReadFrame()
{
  // This should not happen, but just in case....
  if (currentReadFrame == NULL)
    return;

#ifdef JITTER_DEBUG
  if (debugInPos < PARRAYSIZE(debugInTime)) {
    debugInTick[debugInPos] = PTimer::Tick();
    debugInTime[debugInPos++] = currentReadFrame->GetTimestamp();
  }
#endif

  mutex.Wait();

  // Have been reading a frame, put it into the queue now, at correct position
  if (newestFrame == NULL)
    oldestFrame = newestFrame = currentReadFrame;
  else {
    DWORD time = currentReadFrame->GetTimestamp();

    if (time > newestFrame->GetTimestamp()) {
      currentReadFrame->prev = newestFrame;
      newestFrame->next = currentReadFrame;
      newestFrame = currentReadFrame;
    }
    else {
      Entry * frame = newestFrame->prev;
      while (frame != NULL && time > frame->GetTimestamp())
        frame = frame->prev;

      if (frame == NULL) {
        currentReadFrame->next = oldestFrame;
        oldestFrame->prev = currentReadFrame;
        oldestFrame = currentReadFrame;
      }
      else {
        currentReadFrame->prev = frame;
        currentReadFrame->next = frame->next;
        frame->next = currentReadFrame;
        frame->next->prev = currentReadFrame;
      }
    }
  }

  currentReadFrame = NULL;

  mutex.Signal();
}


RTP_DataFrame * RTP_JitterBuffer::GetWriteFrame(DWORD timestamp)
{
  // Get the next frame to write to the codec. Takes it from the oldest
  // position in the queue, if it is time to do so, and parks it in the
  // special member so can unlock the mutex while the writer thread has its
  // way with the buffer.

  if (currentWriteFrame != NULL)
    return currentWriteFrame;

  // No data yet to play!
  if (oldestFrame != NULL) {
    mutex.Wait();

    /* See if time for this packet, if our oldest frame is older than the
       required age, then use it. If it is not time yet, make sure that the
       writer thread isn't falling behind (not enough MIPS). If the time
       between the oldest and the newest entries in the jitter buffer is
       greater than the size specified for the buffer, then return the oldest
       entry regardless, making the writer thread catch up.
    */

    DWORD oldTimestamp = oldestFrame->GetTimestamp();
    DWORD deltaTime = newestFrame->GetTimestamp() - oldTimestamp;

    if (preBuffering && deltaTime > maxJitterTime/2)
      preBuffering = FALSE;

    if (!preBuffering && (oldTimestamp <= timestamp || deltaTime > maxJitterTime)) {
      // Detatch oldest packet from the list, put into parking space
      currentWriteFrame = oldestFrame;
      oldestFrame = currentWriteFrame->next;
      if (oldestFrame != NULL)
        oldestFrame->prev = NULL;
      else
        newestFrame = NULL;
      currentWriteFrame->next = NULL;
    }

    mutex.Signal();
  }

#ifdef JITTER_DEBUG
  if (debugOutPos < PARRAYSIZE(debugOutTime)) {
    debugOutTick[debugOutPos] = PTimer::Tick();
    debugOutTime[debugOutPos++] = currentWriteFrame != NULL ? currentWriteFrame->GetTimestamp() : 0;
  }
#endif

  return currentWriteFrame;
}


void RTP_JitterBuffer::FreeWriteFrame()
{
  // Free the frame just written to codec (or ignored), putting it back into
  // the free list and clearing the parking spot for it.

  // This should not happen, but just in case....
  if (currentWriteFrame == NULL)
    return;

  mutex.Wait();

  // Have been reading a frame, put it into the queue now
  currentWriteFrame->next = freeFrames;
  if (freeFrames != NULL)
    freeFrames->prev = currentWriteFrame;
  freeFrames = currentWriteFrame;

  currentWriteFrame = NULL;

  mutex.Signal();
}


/////////////////////////////////////////////////////////////////////////////
