/*
 * codecs.cxx
 *
 * H.323 protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: codecs.cxx,v $
 * Revision 1.18  1999/08/31 12:34:18  robertj
 * Added gatekeeper support.
 *
 * Revision 1.17  1999/08/25 05:05:36  robertj
 * Added UserInput capability.
 * Allowed the attachment of a channel on a codec to optionally delete the channel object,
 * Improved opening of audio codecs, PSoundChannel creation now in endpoint.
 *
 * Revision 1.16  1999/07/16 16:05:48  robertj
 * Added "human readable" codec type name display.
 *
 * Revision 1.15  1999/07/16 15:01:30  robertj
 * Added message print when starting GSM codec.
 *
 * Revision 1.14  1999/07/15 14:45:36  robertj
 * Added propagation of codec open error to shut down logical channel.
 * Fixed control channel start up bug introduced with tunnelling.
 *
 * Revision 1.13  1999/07/13 09:53:24  robertj
 * Fixed some problems with jitter buffer and added more debugging.
 *
 * Revision 1.12  1999/07/10 02:42:53  robertj
 * Fixed interopability problem with NetMetting 2.1 G711 capability.
 *
 * Revision 1.11  1999/07/09 06:09:49  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.10  1999/06/24 13:32:45  robertj
 * Fixed ability to change sound device on codec and fixed NM3 G.711 compatibility
 *
 * Revision 1.9  1999/06/22 13:49:40  robertj
 * Added GSM support and further RTP protocol enhancements.
 *
 * Revision 1.8  1999/06/14 15:08:40  robertj
 * Added GSM codec class frame work (still no actual codec).
 *
 * Revision 1.7  1999/06/14 08:44:58  robertj
 * Fixed sound buffers to be correct size for stream based audio.
 * GNU C++ compatibility
 *
 * Revision 1.6  1999/06/14 06:39:08  robertj
 * Fixed problem with getting transmit flag to channel from PDU negotiator
 *
 * Revision 1.5  1999/06/14 05:15:55  robertj
 * Changes for using RTP sessions correctly in H323 Logical Channel context
 *
 * Revision 1.4  1999/06/13 12:41:14  robertj
 * Implement logical channel transmitter.
 * Fixed H245 connect on receiving call.
 *
 * Revision 1.3  1999/06/09 05:26:19  robertj
 * Major restructuring of classes.
 *
 * Revision 1.2  1999/06/06 06:06:36  robertj
 * Changes for new ASN compiler and v2 protocol ASN files.
 *
 * Revision 1.1  1999/01/16 01:31:04  robertj
 * Initial revision
 *
 */

#include <ptlib.h>
#include <ptlib/svcproc.h>
#include "codecs.h"
#include "channels.h"
#include "h323pdu.h"


#define new PNEW


/////////////////////////////////////////////////////////////////////////////

BOOL H323SimultaneousCapabilities::SetSize(PINDEX newSize)
{
  PINDEX oldSize = GetSize();

  if (!H323CapabilityListArray::SetSize(newSize))
    return FALSE;

  while (oldSize < newSize) {
    H323CapabilityList * list = new H323CapabilityList;
    // The lowest level list should not delete codecs on destruction
    list->DisallowDeleteObjects();
    SetAt(oldSize++, list);
  }

  return TRUE;
}


BOOL H323CapabilitySet::SetSize(PINDEX newSize)
{
  PINDEX oldSize = GetSize();

  if (!H323CapabilitySetArray::SetSize(newSize))
    return FALSE;

  while (oldSize < newSize)
    SetAt(oldSize++, new H323SimultaneousCapabilities);

  return TRUE;
}


/////////////////////////////////////////////////////////////////////////////

H323Capability::H323Capability()
{
  assignedCapabilityNumber = 1;
}


PObject::Comparison H323Capability::Compare(const PObject & obj) const
{
  PAssert(obj.IsDescendant(H323Capability::Class()), PInvalidCast);
  const H323Capability & other = (const H323Capability &)obj;

  int mt = GetMainType();
  int omt = other.GetMainType();
  if (mt < omt)
    return LessThan;
  if (mt > omt)
    return GreaterThan;

  int st = GetSubType();
  int ost = other.GetSubType();
  if (st < ost)
    return LessThan;
  if (st > ost)
    return GreaterThan;

  return EqualTo;
}


void H323Capability::PrintOn(ostream & strm) const
{
  strm << '<' << assignedCapabilityNumber << '>';
}


unsigned H323Capability::GetDefaultSessionID() const
{
  return 0;
}


void H323Capability::AssignCapabilityCode(const H323CapabilityList & capabilities)
{
  PINDEX i = 0;
  while (i < capabilities.GetSize()) {
    if (capabilities[i].assignedCapabilityNumber != assignedCapabilityNumber)
      i++;
    else {
      // If it already in use, increment it
      assignedCapabilityNumber++;
      i = 0;
    }
  }
}


#ifdef PTRACING
const char * const H323Capability::MainTypesNames[] = {
  "Audio", "Video", "Data", "UserInput"
};
#endif


/////////////////////////////////////////////////////////////////////////////

H323Channel * H323RealTimeCapability::CreateChannel(H323Connection & connection,
                                                    H323Channel::Directions dir,
                                                    unsigned sessionID,
                                 const H245_H2250LogicalChannelParameters * param) const
{
  RTP_Session * session;

  if (param != NULL)
    session = connection.UseSession(param->m_sessionID, param->m_mediaControlChannel);
  else {
    // Make a fake transmprt address from the connection so gets initialised with
    // the transport type (IP, IPX, multicast etc).
    H245_TransportAddress addr;
    connection.GetControlChannel().SetUpTransportPDU(addr, 0);
    session = connection.UseSession(sessionID, addr);
  }

  if (session == NULL)
    return NULL;

  return new H323_RTPChannel(connection, *this, dir, *session);
}


/////////////////////////////////////////////////////////////////////////////

H323Capability::MainTypes H323AudioCapability::GetMainType() const
{
  return e_Audio;
}


BOOL H323AudioCapability::OnSendingPDU(H245_Capability & cap) const
{
  cap.SetTag(H245_Capability::e_receiveAudioCapability);
  return OnSendingPDU((H245_AudioCapability &)cap);
}


BOOL H323AudioCapability::OnSendingPDU(H245_DataType & dataType) const
{
  dataType.SetTag(H245_DataType::e_audioData);
  return OnSendingPDU((H245_AudioCapability &)dataType);
}


BOOL H323AudioCapability::OnReceivedPDU(const H245_Capability & cap)
{
  if (cap.GetTag() != H245_Capability::e_receiveAudioCapability)
    return FALSE;

  return OnReceivedPDU((const H245_AudioCapability &)cap);
}


BOOL H323AudioCapability::OnReceivedPDU(const H245_DataType & dataType)
{
  if (dataType.GetTag() != H245_DataType::e_audioData)
    return FALSE;

  return OnReceivedPDU((const H245_AudioCapability &)dataType);
}


unsigned H323AudioCapability::GetDefaultSessionID() const
{
  return RTP_Session::DefaultAudioSessionID;
}


/////////////////////////////////////////////////////////////////////////////

H323Capability::MainTypes H323VideoCapability::GetMainType() const
{
  return e_Video;
}


BOOL H323VideoCapability::OnSendingPDU(H245_Capability & cap) const
{
  cap.SetTag(H245_Capability::e_receiveVideoCapability);
  return OnSendingPDU((H245_VideoCapability &)cap);
}


BOOL H323VideoCapability::OnSendingPDU(H245_DataType & dataType) const
{
  dataType.SetTag(H245_DataType::e_videoData);
  return OnSendingPDU((H245_VideoCapability &)dataType);
}


BOOL H323VideoCapability::OnReceivedPDU(const H245_Capability & cap)
{
  if (cap.GetTag() != H245_Capability::e_receiveVideoCapability)
    return FALSE;

  return OnReceivedPDU((const H245_VideoCapability &)cap);
}


BOOL H323VideoCapability::OnReceivedPDU(const H245_DataType & dataType)
{
  if (dataType.GetTag() != H245_DataType::e_videoData)
    return FALSE;

  return OnReceivedPDU((const H245_VideoCapability &)dataType);
}


unsigned H323VideoCapability::GetDefaultSessionID() const
{
  return RTP_Session::DefaultVideoSessionID;
}


/////////////////////////////////////////////////////////////////////////////

H323Capability::MainTypes H323DataCapability::GetMainType() const
{
  return e_Data;
}


BOOL H323DataCapability::OnSendingPDU(H245_Capability & cap) const
{
  cap.SetTag(H245_Capability::e_receiveAndTransmitDataApplicationCapability);
  return OnSendingPDU((H245_DataApplicationCapability &)cap);
}


BOOL H323DataCapability::OnSendingPDU(H245_DataType & dataType) const
{
  dataType.SetTag(H245_DataType::e_data);
  return OnSendingPDU((H245_DataApplicationCapability &)dataType);
}


BOOL H323DataCapability::OnReceivedPDU(const H245_Capability & cap)
{
  if (cap.GetTag() != H245_Capability::e_receiveDataApplicationCapability)
    return FALSE;

  return OnReceivedPDU((const H245_DataApplicationCapability &)cap);
}


BOOL H323DataCapability::OnReceivedPDU(const H245_DataType & dataType)
{
  if (dataType.GetTag() != H245_DataType::e_data)
    return FALSE;

  return OnReceivedPDU((const H245_DataApplicationCapability &)dataType);
}


H323Codec * H323DataCapability::CreateCodec(H323Codec::Direction) const
{
  return NULL;
}


/////////////////////////////////////////////////////////////////////////////

H323Codec::H323Codec(Direction dir)
{
  direction = dir;
  rtpPayloadType = RTP_DataFrame::DynamicBase;
}


BOOL H323Codec::Open(H323Connection & /*connection*/)
{
  return TRUE;
}


void H323Codec::Close()
{
}


/////////////////////////////////////////////////////////////////////////////

H323AudioCodec::H323AudioCodec(Direction dir, unsigned samples)
  : H323Codec(dir)
{
  audio = NULL;
  deleteAudioChannel = FALSE;
  samplesPerFrame = samples;
}


H323AudioCodec::~H323AudioCodec()
{
  if (deleteAudioChannel)
    delete audio;
}


BOOL H323AudioCodec::Open(H323Connection & connection)
{
  audio = connection.GetEndPoint().OpenAudioChannel(direction == Encoder, samplesPerFrame*2);
  return audio != NULL;
}


void H323AudioCodec::Close()
{
  if (audio != NULL)
    audio->Close();
}


unsigned H323AudioCodec::GetFrameRate() const
{
  return samplesPerFrame;
}


unsigned H323AudioCodec::GetTimeUnits() const
{
  return 8; // Assuming standard 8kHz sample rate
}


void H323AudioCodec::AttachAudioChannel(PChannel * channel, BOOL autoDelete)
{
  if (deleteAudioChannel)
    delete audio;
  audio = channel;
  deleteAudioChannel = autoDelete;
}


/////////////////////////////////////////////////////////////////////////////

H323FramedAudioCodec::H323FramedAudioCodec(Direction dir,
                                           unsigned samples,
                                           unsigned bytes)
  : H323AudioCodec(dir, samples),
    sampleBuffer(samples)
{
  bytesPerFrame = bytes;
}


BOOL H323FramedAudioCodec::Read(BYTE * buffer, unsigned & length)
{
  if (audio == NULL) {
    PTRACE(1, "Codec\tNo audio channel for read");
    return FALSE;
  }

  if (direction != Encoder) {
    PTRACE(1, "Codec\tAttempt to decode from encoder");
    return FALSE;
  }

  if (!audio->Read(sampleBuffer.GetPointer(samplesPerFrame), samplesPerFrame*2)) {
    PTRACE(1, "Codec\tAudio read failed: " << audio->GetErrorText());
    return FALSE;
  }

  length = bytesPerFrame;

  return EncodeFrame(buffer);
}


BOOL H323FramedAudioCodec::Write(const BYTE * buffer, unsigned length, unsigned & written)
{
  if (audio == NULL) {
    PTRACE(1, "Codec\tNo audio channel for write");
    return FALSE;
  }

  if (direction != Decoder) {
    PTRACE(1, "Codec\tAttempt to encode from decoder");
    return FALSE;
  }

  // If length is zero then it indicates silence, do nothing.
  if (length == 0) {
    written = 0;
    return TRUE;
  }

  if (length < bytesPerFrame) {
    /*Need to add something in here if not enough bytes (ie data in original
      packet was not evenly divisible by bytesPerFrame) to save the partial
      frame till the next call to Write(). Theoretically this is not
      necessary as the H323Capability should dictate the quanta to be used
      but may end up being needed for some codec types where it is supposed
      to be streamed but due to hardware is framed.

      At present this is an error condition.
     */
    return FALSE;
  }

  // Decode the data
  if (!DecodeFrame(buffer))
    return FALSE;

  // Write as 16bit PCM to sound channel
  if (!audio->Write(sampleBuffer, samplesPerFrame*2))
    PTRACE(1, "Codec\tWrite failed: " << audio->GetErrorText());

  written = bytesPerFrame;

  return TRUE;
}


unsigned H323FramedAudioCodec::GetMaxFrameSize() const
{
  return bytesPerFrame;
}


unsigned H323FramedAudioCodec::GetBandwidth() const
{
  // Want 100's of bits per second from the number of bytes in a frame
  // the number of time units per frame and number of units per millisecond.
  return 80*bytesPerFrame*GetTimeUnits()/GetFrameRate();
}


/////////////////////////////////////////////////////////////////////////////

H323StreamedAudioCodec::H323StreamedAudioCodec(Direction dir,
                                               unsigned samples,
                                               unsigned bits)
  : H323FramedAudioCodec(dir, samples, (samples*bits+7)/8)
{
  bitsPerSample = bits;
}


BOOL H323StreamedAudioCodec::EncodeFrame(BYTE * buffer)
{
  PINDEX i;

  switch (bitsPerSample) {
    case 8 :
      for (i = 0; i < (PINDEX)samplesPerFrame; i++)
        *buffer++ = (BYTE)Encode(sampleBuffer[i]);
      break;

    case 4 :
      for (i = 0; i < (PINDEX)samplesPerFrame; i++) {
        if ((i&1) == 0)
          *buffer = (BYTE)Encode(sampleBuffer[i]);
        else
          *buffer |= (BYTE)(Encode(sampleBuffer[i]) << 4);
      }
      break;

    default :
      PAssertAlways("Unsupported bit size");
      return FALSE;
  }

  return TRUE;
}


BOOL H323StreamedAudioCodec::DecodeFrame(const BYTE * buffer)
{
  PINDEX i;
  short * out = sampleBuffer.GetPointer(samplesPerFrame);

  switch (bitsPerSample) {
    case 8 :
      for (i = 0; i < (PINDEX)samplesPerFrame; i++)
        *out++ = Decode(*buffer++);
      break;

    case 4 :
      for (i = 0; i < (PINDEX)samplesPerFrame; i++) {
        *out++ = Decode(*buffer & 15);
        *out++ = Decode(*buffer >> 4);
        buffer++;
      }
      break;

    default :
      PAssertAlways("Unsupported bit size");
      return FALSE;
  }

  return TRUE;
}


/////////////////////////////////////////////////////////////////////////////

H323_G711Capability::H323_G711Capability()
{
  mode = muLaw;
  speed = At64k;
  frameSize = 30;
}


H323_G711Capability::H323_G711Capability(Mode m, Speed s)
{
  mode = m;
  speed = s;
  frameSize = 30;
}


PObject * H323_G711Capability::Clone() const
{
  return new H323_G711Capability(mode, speed);
}


void H323_G711Capability::PrintOn(ostream & strm) const
{
  strm << "G.711 "
       << (mode == muLaw ? "u-Law " : "A-Law ")
       << (speed == At64k ? "64k " : "56k ");
  H323AudioCapability::PrintOn(strm);
}


static const unsigned G711SubType[2][2] = {
  { H245_AudioCapability::e_g711Alaw64k, H245_AudioCapability::e_g711Alaw56k },
  { H245_AudioCapability::e_g711Ulaw64k, H245_AudioCapability::e_g711Ulaw56k }
};


unsigned H323_G711Capability::GetSubType() const
{
  return G711SubType[mode][speed];
}


BOOL H323_G711Capability::OnSendingPDU(H245_AudioCapability & cap) const
{
  cap.SetTag(G711SubType[mode][speed]);

  PASN_Integer & value = cap;
  // This is supposed to be the maximum number of audio frames per packet, but
  // what that means in the context of a stream based codec such as G.711 is
  // not well defined. We believe this to be milliseconds and have set it
  // accordinglingly.
  value = frameSize;
  return TRUE;
}


BOOL H323_G711Capability::OnReceivedPDU(const H245_AudioCapability & cap)
{
  switch (cap.GetTag()) {
    case H245_AudioCapability::e_g711Alaw64k :
      mode = H323_G711Capability::ALaw;
      speed = H323_G711Capability::At64k;
      break;

    case H245_AudioCapability::e_g711Alaw56k :
      mode = H323_G711Capability::ALaw;
      speed = H323_G711Capability::At56k;
      break;

    case H245_AudioCapability::e_g711Ulaw64k :
      mode = H323_G711Capability::muLaw;
      speed = H323_G711Capability::At64k;
      break;

    case H245_AudioCapability::e_g711Ulaw56k :
      mode = H323_G711Capability::muLaw;
      speed = H323_G711Capability::At56k;
      break;

    default :
      return FALSE;
  }

  const PASN_Integer & value = cap;

  /*This is supposed to be the maximum number of audio frames per packet, but
    what that means in the context of a stream based codec such as G.711 is
    not well defined. We believe this to be milliseconds and have set it
    accordinglingly.
   */
  frameSize = value;

  /*Except NetMeeting 2.1 seems to send it as number of samples (sends 240), so
    to help with interoperability we only allow up to 30 milliseconds to be
    sent. This should work with old and new stacks.
   */
  if (frameSize > 30)
    frameSize = 30;

  return TRUE;
}


H323Codec * H323_G711Capability::CreateCodec(H323Codec::Direction direction) const
{
  if (mode == muLaw)
    return new H323_muLawCodec(direction, speed, frameSize);
  else
    return new H323_ALawCodec(direction, speed, frameSize);
}


/////////////////////////////////////////////////////////////////////////////

H323_ALawCodec::H323_ALawCodec(Direction dir,
                               H323_G711Capability::Speed rate,
                               unsigned frameSize)
  : H323StreamedAudioCodec(dir, frameSize*8, 8)
{
  rtpPayloadType = RTP_DataFrame::PCMA;
  sevenBit = rate == H323_G711Capability::At56k;

  PTRACE(3, "Codec\tG711 ALaw " << (dir == Encoder ? "en" : "de")
         << "coder created for at "
         << (sevenBit ? "56k" : "64k") << ", " << frameSize << "ms frames");
}


extern "C" unsigned char linear2alaw(int pcm_val);

int H323_ALawCodec::Encode(short sample) const
{
  return linear2alaw(sample);
}


extern "C" int alaw2linear(unsigned char u_val);

short H323_ALawCodec::Decode(int sample) const
{
  return (short)alaw2linear((unsigned char)sample);
}


/////////////////////////////////////////////////////////////////////////////

H323_muLawCodec::H323_muLawCodec(Direction dir,
                                 H323_G711Capability::Speed rate,
                                 unsigned frameSize)
  : H323StreamedAudioCodec(dir, frameSize*8, 8)
{
  rtpPayloadType = RTP_DataFrame::PCMU;
  sevenBit = rate == H323_G711Capability::At56k;
  PTRACE(3, "Codec\tG711 uLaw codec created at "
         << (sevenBit ? "56k" : "64k") << ", frame of " << frameSize << "ms");
}


extern "C" unsigned char linear2ulaw(int pcm_val);

int H323_muLawCodec::Encode(short sample) const
{
  return linear2ulaw(sample);
}


extern "C" int ulaw2linear(unsigned char u_val);

short H323_muLawCodec::Decode(int sample) const
{
  return (short)ulaw2linear((unsigned char)sample);
}


/////////////////////////////////////////////////////////////////////////////

H323_GSM0610Capability::H323_GSM0610Capability()
{
}


PObject * H323_GSM0610Capability::Clone() const
{
  return new H323_G711Capability();
}


void H323_GSM0610Capability::PrintOn(ostream & strm) const
{
  strm << "GSM 06.10 ";
  H323AudioCapability::PrintOn(strm);
}


unsigned H323_GSM0610Capability::GetSubType() const
{
  return H245_AudioCapability::e_gsmFullRate;
}


BOOL H323_GSM0610Capability::OnSendingPDU(H245_AudioCapability & cap) const
{
  cap.SetTag(H245_AudioCapability::e_gsmFullRate);

  H245_GSMAudioCapability & gsm = cap;
  gsm.m_audioUnitSize = H323_GSM0610Codec::SamplesPerFrame*5;  // 100 milliseconds
  return TRUE;
}


BOOL H323_GSM0610Capability::OnReceivedPDU(const H245_AudioCapability & cap)
{
  if (cap.GetTag() != H245_AudioCapability::e_gsmFullRate)
    return FALSE;

//  H245_GSMAudioCapability & gsm = cap;
  // May need to check gsm.m_audioUnitSize at some point, at present codec only
  // sends one frame at a time anyway
  return TRUE;
}


H323Codec * H323_GSM0610Capability::CreateCodec(H323Codec::Direction direction) const
{
  return new H323_GSM0610Codec(direction);
}


/////////////////////////////////////////////////////////////////////////////

extern "C" gsm_state * gsm_create();
//gsm_state * gsm_create() { return NULL; }

H323_GSM0610Codec::H323_GSM0610Codec(Direction dir)
  : H323FramedAudioCodec(dir, SamplesPerFrame, BytesPerFrame)
{
  rtpPayloadType = RTP_DataFrame::GSM;
  gsm = gsm_create();
  PTRACE(3, "Codec\tGSM " << (dir == Encoder ? "en" : "de")
         << "coder created");
}


extern "C" void gsm_destroy(gsm_state *);
//void gsm_destroy(gsm_state *) { }

H323_GSM0610Codec::~H323_GSM0610Codec()
{
  gsm_destroy(gsm);
}


extern "C" void gsm_encode(gsm_state *, const short *, BYTE *);
//void gsm_encode(gsm_state *, const short *, BYTE *) { }

BOOL H323_GSM0610Codec::EncodeFrame(BYTE * buffer)
{
  gsm_encode(gsm, sampleBuffer, buffer);
  return TRUE;
}


extern "C" void gsm_decode(gsm_state *, const BYTE *, short *);
//void gsm_decode(gsm_state *, const BYTE *, short *) { }

BOOL H323_GSM0610Codec::DecodeFrame(const BYTE * buffer)
{
  gsm_decode(gsm, buffer, sampleBuffer.GetPointer());
  return TRUE;
}


/////////////////////////////////////////////////////////////////////////////

PObject * H323_T120Capability::Clone() const
{
  return new H323_T120Capability();
}


unsigned H323_T120Capability::GetSubType() const
{
  return H245_DataApplicationCapability_application::e_t120;
}


BOOL H323_T120Capability::OnSendingPDU(H245_DataApplicationCapability & cap) const
{
  cap.m_maxBitRate = 85000;
  cap.m_application.SetTag(H245_DataApplicationCapability_application::e_t120);
  H245_DataProtocolCapability & proto = cap.m_application;
  proto.SetTag(H245_DataProtocolCapability::e_separateLANStack);
  return TRUE;
}


BOOL H323_T120Capability::OnReceivedPDU(const H245_DataApplicationCapability & cap)
{
  if (cap.m_application.GetTag() != H245_DataApplicationCapability_application::e_t120)
    return FALSE;

  const H245_DataProtocolCapability & dataCap = cap.m_application;

  return dataCap.GetTag() == H245_DataProtocolCapability::e_separateLANStack;
}


H323Channel * H323_T120Capability::CreateChannel(H323Connection &,
                                                 H323Channel::Directions,
                                                 unsigned,
                                const H245_H2250LogicalChannelParameters *) const
{
  return NULL;
}


/////////////////////////////////////////////////////////////////////////////

H323_UserInputCapability::H323_UserInputCapability(unsigned _subType)
{
  subType = _subType;
}


PObject * H323_UserInputCapability::Clone() const
{
  return new H323_UserInputCapability(subType);
}


void H323_UserInputCapability::PrintOn(ostream & strm) const
{
  switch (subType) {
    case H245_UserInputCapability::e_basicString :
      strm << "basicString";
      break;
    case H245_UserInputCapability::e_iA5String :
      strm << "iA5String";
      break;
    case H245_UserInputCapability::e_generalString :
      strm << "generalString";
      break;
    case H245_UserInputCapability::e_dtmf :
      strm << "dtmf";
      break;
    case H245_UserInputCapability::e_hookflash :
      strm << "hookflash";
      break;
    default :
      strm << "SubType(" << subType << ')';
  }
  H323Capability::PrintOn(strm);
}


H323Capability::MainTypes H323_UserInputCapability::GetMainType() const
{
  return e_UserInput;
}


unsigned  H323_UserInputCapability::GetSubType()  const
{
  return subType;
}


BOOL H323_UserInputCapability::OnSendingPDU(H245_Capability & pdu) const
{
  pdu.SetTag(H245_Capability::e_receiveUserInputCapability);
  H245_UserInputCapability & ui = pdu;
  ui.SetTag(subType);
  return TRUE;
}


BOOL H323_UserInputCapability::OnSendingPDU(H245_DataType &) const
{
  PTRACE(1, "Codec\tCannot have UserInputCapability in DataType");
  return FALSE;
}


BOOL H323_UserInputCapability::OnReceivedPDU(const H245_Capability & pdu)
{
  if (pdu.GetTag() != H245_Capability::e_receiveUserInputCapability)
    return FALSE;

  const H245_UserInputCapability & ui = pdu;
  subType = ui.GetTag();
  return TRUE;
}


BOOL H323_UserInputCapability::OnReceivedPDU(const H245_DataType &)
{
  PTRACE(1, "Codec\tCannot have UserInputCapability in DataType");
  return FALSE;
}


H323Channel * H323_UserInputCapability::CreateChannel(H323Connection &,
                                                      H323Channel::Directions,
                                                      unsigned,
                                                      const H245_H2250LogicalChannelParameters *) const
{
  PTRACE(1, "Codec\tCannot create UserInputCapability channel");
  return NULL;
}


H323Codec * H323_UserInputCapability::CreateCodec(H323Codec::Direction) const
{
  PTRACE(1, "Codec\tCannot create UserInputCapability codec");
  return NULL;
}


/////////////////////////////////////////////////////////////////////////////
