/*
 * gkclient.cxx
 *
 * Gatekeeper client protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: gkclient.cxx,v $
 * Revision 1.1  1999/08/31 12:34:18  robertj
 * Added gatekeeper support.
 *
 */

#include <ptlib.h>
#include "gkclient.h"
#include "h323pdu.h"

class H225RasThread : public PThread
{
  PCLASSINFO(H225RasThread, PThread)

  public:
    H225RasThread(H323EndPoint & endpoint,
                  H323Gatekeeper & gatekeeper);

  protected:
    void Main();

    H323Gatekeeper & gatekeeper;
};


#define new PNEW


/////////////////////////////////////////////////////////////////////////////

H225RasThread::H225RasThread(H323EndPoint & endpoint,
                             H323Gatekeeper & g)
  : PThread(endpoint.GetRasThreadStackSize(), NoAutoDeleteThread),
    gatekeeper(g)
{
  Resume();
}


void H225RasThread::Main()
{
  PTRACE(3, "H225\tStarted RAS thread");

  gatekeeper.HandleRasChannel();
}


/////////////////////////////////////////////////////////////////////////////

H323Gatekeeper::H323Gatekeeper(H323EndPoint & ep, H323Transport * trans)
  : endpoint(ep),
    transport(trans)
{
  lastSequenceNumber = 0;
  isRegistered = FALSE;
  allocatedBandwidth = 0;
}


H323Gatekeeper::~H323Gatekeeper()
{
  delete transport;
}


void H323Gatekeeper::PrintOn(ostream & strm) const
{
  if (!gatekeeperIdentifier)
    strm << gatekeeperIdentifier << "@";
  strm << transport->GetRemoteAddress();
}


void H323Gatekeeper::HandleRasChannel()
{
  transport->SetReadTimeout(PMaxTimeInterval);

  H323RasPDU response;
  while (response.Read(*transport)) {
    if (HandleRasPDU(response))
      responseHandled.Signal();
  }
}


BOOL H323Gatekeeper::HandleRasPDU(const H323RasPDU & response)
{
  switch (response.GetTag()) {
    case H225_RasMessage::e_gatekeeperRequest :
      OnReceiveGatekeeperRequest(response);
      break;

    case H225_RasMessage::e_gatekeeperConfirm :
      return OnReceiveGatekeeperConfirm(response);

    case H225_RasMessage::e_gatekeeperReject :
      return OnReceiveGatekeeperReject(response);

    case H225_RasMessage::e_registrationRequest :
      OnReceiveRegistrationRequest(response);
      break;

    case H225_RasMessage::e_registrationConfirm :
      return OnReceiveRegistrationConfirm(response);

    case H225_RasMessage::e_registrationReject :
      return OnReceiveRegistrationReject(response);

    case H225_RasMessage::e_unregistrationRequest :
      OnReceiveUnregistrationRequest(response);
      break;

    case H225_RasMessage::e_unregistrationConfirm :
      return OnReceiveUnregistrationConfirm(response);

    case H225_RasMessage::e_unregistrationReject :
      return OnReceiveUnregistrationReject(response);

    case H225_RasMessage::e_admissionRequest :
      OnReceiveAdmissionRequest(response);
      break;

    case H225_RasMessage::e_admissionConfirm :
      return OnReceiveAdmissionConfirm(response);

    case H225_RasMessage::e_admissionReject :
      return OnReceiveAdmissionReject(response);

    case H225_RasMessage::e_bandwidthRequest :
      OnReceiveBandwidthRequest(response);
      break;

    case H225_RasMessage::e_bandwidthConfirm :
      return OnReceiveBandwidthConfirm(response);

    case H225_RasMessage::e_bandwidthReject :
      return OnReceiveBandwidthReject(response);

    case H225_RasMessage::e_disengageRequest :
      OnReceiveDisengageRequest(response);
      break;

    case H225_RasMessage::e_disengageConfirm :
      return OnReceiveDisengageConfirm(response);

    case H225_RasMessage::e_disengageReject :
      return OnReceiveDisengageReject(response);

    case H225_RasMessage::e_locationRequest :
      OnReceiveLocationRequest(response);
      break;

    case H225_RasMessage::e_locationConfirm :
      return OnReceiveLocationConfirm(response);

    case H225_RasMessage::e_locationReject :
      return OnReceiveLocationReject(response);

    case H225_RasMessage::e_infoRequest :
      OnReceiveInfoRequest(response);
      break;

    case H225_RasMessage::e_infoRequestResponse :
      return OnReceiveInfoRequestResponse(response);

    case H225_RasMessage::e_nonStandardMessage :
      OnReceiveNonStandardMessage(response);
      break;

    case H225_RasMessage::e_unknownMessageResponse :
      OnReceiveUnknownMessageResponse(response);
      break;

    case H225_RasMessage::e_requestInProgress :
      OnReceiveRequestInProgress(response);
      break;

    case H225_RasMessage::e_resourcesAvailableIndicate :
      OnReceiveResourcesAvailableIndicate(response);
      break;

    case H225_RasMessage::e_resourcesAvailableConfirm :
      return OnReceiveResourcesAvailableConfirm(response);

    case H225_RasMessage::e_infoRequestAck :
      return OnReceiveInfoRequestAck(response);

    case H225_RasMessage::e_infoRequestNak :
      return OnReceiveInfoRequestNak(response);

    default :
      OnReceiveUnkown(response);
  }

  return FALSE;
}


BOOL H323Gatekeeper::MakeRequest(const H323RasPDU & request)
{
  PTRACE(3, "RAS\tMaking request: " << request.GetTagName());

  // Assume a confirm, reject callbacks will set to RejectReceived
  responseResult = ConfirmReceived;

  // Set request code for determining response
  requestTag = request.GetTag();

  unsigned retries = endpoint.GetRasRequestRetries();
  while (retries-- > 0) {
    if (!request.Write(*transport))
      return FALSE;

    while (responseHandled.Wait(endpoint.GetRasRequestTimeout())) {
      switch (responseResult) {
        case ConfirmReceived :
          return TRUE;

        case RejectReceived :
          return FALSE;

        case RequestInProgress :
          responseResult = ConfirmReceived;
          // Keep waiting
      }
    }

    PTRACE(2, "RAS\tTimeout on request " << retries);
  }

  return FALSE;
}


BOOL H323Gatekeeper::CheckForResponse(unsigned reqTag, unsigned seqNum)
{
  if (requestTag != reqTag)
    return FALSE;

  if (seqNum == lastSequenceNumber)
    return TRUE;

  PTRACE(3, "RAS\tReceived incorrect sequence number, want "
         << lastSequenceNumber << ", got " << seqNum);
  return FALSE;
}


BOOL H323Gatekeeper::OnReceiveRequestInProgress(const H225_RequestInProgress & rip)
{
  if (!CheckForResponse(requestTag, rip.m_requestSeqNum))
    return FALSE;

  responseResult = RequestInProgress;

  return TRUE;
}


BOOL H323Gatekeeper::Discover()
{
  H323RasPDU request;
  H225_GatekeeperRequest & grq = request.BuildGatekeeperRequest(++lastSequenceNumber);
  endpoint.SetEndpointTypeInfo(grq.m_endpointType);
  transport->SetUpTransportPDU(grq.m_rasAddress, TRUE);

  OnSendGatekeeperRequest(grq);

  unsigned retries = endpoint.GetGatekeeperRequestRetries();
  transport->SetReadTimeout(endpoint.GetGatekeeperRequestTimeout());
  requestTag = request.GetTag();

  while (!transport->DiscoverGatekeepers(*this, request)) {
    if (--retries == 0)
      return FALSE;
  }

  if (transport->Connect())
    transport->AttachThread(new H225RasThread(endpoint, *this));

  return TRUE;
}


BOOL H323Gatekeeper::OnReceiveGatekeeperConfirm(const H225_GatekeeperConfirm & gcf)
{
  if (!CheckForResponse(H225_RasMessage::e_gatekeeperRequest, gcf.m_requestSeqNum))
    return FALSE;

  locatedAddress = gcf.m_rasAddress;
  PTRACE(2, "H225\tGatekeeper discovery found " << locatedAddress);

  if (!transport->SetRemoteAddress(locatedAddress))
    PTRACE(2, "H225\tInvalid gatekeeper discovery address: \"" << locatedAddress << '"');

  return TRUE;
}


BOOL H323Gatekeeper::RegistrationRequest(const H323TransportAddress & address)
{
  if (!address)
    transport->AttachThread(new H225RasThread(endpoint, *this));

  H323RasPDU request;
  H225_RegistrationRequest & rrq = request.BuildRegistrationRequest(++lastSequenceNumber);
  rrq.m_discoveryComplete = address.IsEmpty();

  const H323ListenerList & listeners = endpoint.GetListeners();
  for (PINDEX i = 0; i < listeners.GetSize(); i++)
    listeners[i].SetUpTransportPDU(rrq.m_callSignalAddress);

  rrq.m_rasAddress.SetSize(1);
  transport->SetUpTransportPDU(rrq.m_rasAddress[0], TRUE);

  endpoint.SetEndpointTypeInfo(rrq.m_terminalType);
  endpoint.SetVendorIdentifierInfo(rrq.m_endpointVendor);

  if (endpoint.IsTerminal()) {
    rrq.IncludeOptionalField(H225_RegistrationRequest::e_terminalAlias);
    rrq.m_terminalAlias.SetSize(1);
    H323SetAliasAddress(endpoint.GetLocalUserName(), rrq.m_terminalAlias[0]);
  }

  if (!gatekeeperIdentifier) {
    rrq.IncludeOptionalField(H225_RegistrationRequest::e_gatekeeperIdentifier);
    rrq.m_gatekeeperIdentifier = gatekeeperIdentifier;
  }

  if (!endpointIdentifier) {
    rrq.IncludeOptionalField(H225_RegistrationRequest::e_endpointIdentifier);
    rrq.m_endpointIdentifier = endpointIdentifier;
  }

  OnSendRegistrationRequest(rrq);

  PWaitAndSignal mutex(makeRequestMutex);

  return MakeRequest(request);
}


BOOL H323Gatekeeper::OnReceiveRegistrationConfirm(const H225_RegistrationConfirm & rcf)
{
  if (!CheckForResponse(H225_RasMessage::e_registrationRequest, rcf.m_requestSeqNum))
    return FALSE;

  isRegistered = TRUE;

  gatekeeperIdentifier = rcf.m_gatekeeperIdentifier;
  endpointIdentifier = rcf.m_endpointIdentifier;

  if (rcf.HasOptionalField(H225_RegistrationConfirm::e_timeToLive))
    timeToLive = PTimeInterval(0, rcf.m_timeToLive);
  else {
    timeToLive = PMaxTimeInterval;
    timeToLive.Stop();
  }

  return TRUE;
}


BOOL H323Gatekeeper::UnregistrationRequest(int reason)
{
  H323RasPDU request;
  H225_UnregistrationRequest & urq = request.BuildUnregistrationRequest(++lastSequenceNumber);

  const H323ListenerList & listeners = endpoint.GetListeners();
  for (PINDEX i = 0; i < listeners.GetSize(); i++)
    listeners[i].SetUpTransportPDU(urq.m_callSignalAddress);

  if (endpoint.IsTerminal()) {
    urq.IncludeOptionalField(H225_UnregistrationRequest::e_endpointAlias);
    urq.m_endpointAlias.SetSize(1);
    H323SetAliasAddress(endpoint.GetLocalUserName(), urq.m_endpointAlias[0]);
  }

  if (!gatekeeperIdentifier) {
    urq.IncludeOptionalField(H225_UnregistrationRequest::e_gatekeeperIdentifier);
    urq.m_gatekeeperIdentifier = gatekeeperIdentifier;
  }

  if (!endpointIdentifier) {
    urq.IncludeOptionalField(H225_UnregistrationRequest::e_endpointIdentifier);
    urq.m_endpointIdentifier = endpointIdentifier;
  }

  if (reason >= 0) {
    urq.IncludeOptionalField(H225_UnregistrationRequest::e_reason);
    urq.m_reason = reason;
  }

  OnSendUnregistrationRequest(urq);

  PWaitAndSignal mutex(makeRequestMutex);

  return MakeRequest(request);
}


BOOL H323Gatekeeper::OnReceiveUnregistrationConfirm(const H225_UnregistrationConfirm & ucf)
{
  if (!CheckForResponse(H225_RasMessage::e_unregistrationRequest, ucf.m_requestSeqNum))
    return FALSE;

  isRegistered = FALSE;

  return TRUE;
}


BOOL H323Gatekeeper::OnReceiveUnregistrationRequest(const H225_UnregistrationRequest & urq)
{
  PTRACE(2, "RAS\tUnregistration received");
  endpoint.ClearAllCalls(TRUE);

  H323RasPDU response;
  response.BuildUnregistrationConfirm(urq.m_requestSeqNum);
  return response.Write(*transport);
}


BOOL H323Gatekeeper::LocationRequest(const PString & alias,
                                     H323TransportAddress & address)
{
  PStringList aliases;
  aliases.AppendString(alias);
  return LocationRequest(aliases, address);
}


BOOL H323Gatekeeper::LocationRequest(const PStringList & aliases,
                                     H323TransportAddress & address)
{
  H323RasPDU request;
  H225_LocationRequest & lrq = request.BuildLocationRequest(++lastSequenceNumber);

  lrq.m_destinationInfo.SetSize(aliases.GetSize());
  for (PINDEX i = 0; i < aliases.GetSize(); i++)
    H323SetAliasAddress(aliases[i], lrq.m_destinationInfo[i]);

  if (!endpointIdentifier) {
    lrq.IncludeOptionalField(H225_LocationRequest::e_endpointIdentifier);
    lrq.m_endpointIdentifier = endpointIdentifier;
  }

  OnSendLocationRequest(lrq);

  PWaitAndSignal mutex(makeRequestMutex);

  if (!MakeRequest(request))
    return FALSE;

  address = locatedAddress;

  return TRUE;
}


BOOL H323Gatekeeper::OnReceiveLocationConfirm(const H225_LocationConfirm & lcf)
{
  if (!CheckForResponse(H225_RasMessage::e_locationRequest, lcf.m_requestSeqNum))
    return FALSE;

  locatedAddress = lcf.m_callSignalAddress;

  return TRUE;
}


BOOL H323Gatekeeper::AdmissionRequest(H323Connection & connection)
{
  H323TransportAddress address;
  return AdmissionRequest(connection, address);
}


BOOL H323Gatekeeper::AdmissionRequest(H323Connection & connection,
                                      H323TransportAddress & address)
{
  H323RasPDU request;
  H225_AdmissionRequest & arq = request.BuildAdmissionRequest(++lastSequenceNumber);


  arq.m_callType.SetTag(H225_CallType::e_pointToPoint);
  arq.m_endpointIdentifier = endpointIdentifier;

  BOOL answeringCall = connection.HadAnsweredCall();
  arq.m_answerCall = answeringCall;

  arq.m_srcInfo.SetSize(1);
  if (answeringCall)
    H323SetAliasAddress(connection.GetRemotePartyName(), arq.m_srcInfo[0]);
  else {
    H323SetAliasAddress(endpoint.GetLocalUserName(), arq.m_srcInfo[0]);
    arq.IncludeOptionalField(H225_AdmissionRequest::e_destinationInfo);
    arq.m_destinationInfo.SetSize(1);
    H323SetAliasAddress(connection.GetRemotePartyName(), arq.m_destinationInfo[0]);
  }

  const H323Transport * signallingChannel = connection.GetSignallingChannel();
  arq.IncludeOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress);
  signallingChannel->SetUpTransportPDU(arq.m_srcCallSignalAddress, !answeringCall);
  arq.IncludeOptionalField(H225_AdmissionRequest::e_destCallSignalAddress);
  signallingChannel->SetUpTransportPDU(arq.m_destCallSignalAddress, answeringCall);

  arq.m_bandWidth = connection.GetBandwidthAvailable();
  arq.m_callReferenceValue = connection.GetCallReference();
  arq.m_conferenceID = connection.GetConferenceIdentifier();
  arq.m_callIdentifier.m_guid = connection.GetCallIdentifier();

  OnSendAdmissionRequest(arq);

  PWaitAndSignal mutex(makeRequestMutex);

  if (!MakeRequest(request))
    return FALSE;

  connection.SetBandwidthAvailable(allocatedBandwidth);
  address = locatedAddress;

  return TRUE;
}


BOOL H323Gatekeeper::OnReceiveAdmissionConfirm(const H225_AdmissionConfirm & acf)
{
  if (!CheckForResponse(H225_RasMessage::e_admissionRequest, acf.m_requestSeqNum))
    return FALSE;

  if (acf.m_callModel.GetTag() == H225_CallModel::e_direct)
    locatedAddress = "";
  else
    locatedAddress = acf.m_destCallSignalAddress;

  allocatedBandwidth = acf.m_bandWidth;

  return TRUE;
}


BOOL H323Gatekeeper::DisengageRequest(const H323Connection & connection, unsigned reason)
{
  H323RasPDU request;
  H225_DisengageRequest & drq = request.BuildDisengageRequest(++lastSequenceNumber);

  drq.m_endpointIdentifier = endpointIdentifier;
  drq.m_conferenceID = connection.GetConferenceIdentifier();
  drq.m_callReferenceValue = connection.GetCallReference();
  drq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
  drq.m_disengageReason.SetTag(reason);
  drq.m_answeredCall = connection.HadAnsweredCall();

  if (!gatekeeperIdentifier) {
    drq.IncludeOptionalField(H225_DisengageRequest::e_gatekeeperIdentifier);
    drq.m_gatekeeperIdentifier = gatekeeperIdentifier;
  }

  OnSendDisengageRequest(drq);

  PWaitAndSignal mutex(makeRequestMutex);

  return MakeRequest(request);
}


BOOL H323Gatekeeper::OnReceiveDisengageConfirm(const H225_DisengageConfirm & dcf)
{
  return CheckForResponse(H225_RasMessage::e_disengageRequest, dcf.m_requestSeqNum);
}


BOOL H323Gatekeeper::OnReceiveDisengageRequest(const H225_DisengageRequest & drq)
{
  H323GloballyUniqueID id = drq.m_callIdentifier.m_guid;
  endpoint.ClearCall(id.AsString(), TRUE);

  H323RasPDU response;
  response.BuildDisengageConfirm(drq.m_requestSeqNum);
  return response.Write(*transport);
}


BOOL H323Gatekeeper::BandwidthRequest(H323Connection & connection,
                                      unsigned requestedBandwidth)
{
  H323RasPDU request;
  H225_BandwidthRequest & brq = request.BuildBandwidthRequest(++lastSequenceNumber);

  brq.m_endpointIdentifier = endpointIdentifier;
  brq.m_conferenceID = connection.GetConferenceIdentifier();
  brq.m_callReferenceValue = connection.GetCallReference();
  brq.m_callIdentifier.m_guid = connection.GetCallIdentifier();
  brq.m_bandWidth = requestedBandwidth;

  OnSendBandwidthRequest(brq);

  PWaitAndSignal mutex(makeRequestMutex);

  if (!MakeRequest(request))
    return FALSE;

  connection.SetBandwidthAvailable(allocatedBandwidth);

  return TRUE;
}


BOOL H323Gatekeeper::OnReceiveBandwidthConfirm(const H225_BandwidthConfirm & bcf)
{
  if (!CheckForResponse(H225_RasMessage::e_bandwidthRequest, bcf.m_requestSeqNum))
    return FALSE;

  allocatedBandwidth = bcf.m_bandWidth;

  return TRUE;
}


BOOL H323Gatekeeper::OnReceiveBandwidthRequest(const H225_BandwidthRequest & brq)
{
  H323GloballyUniqueID id = brq.m_callIdentifier.m_guid;
  H323Connection * connection = endpoint.FindConnectionWithLock(id.AsString());

  H323RasPDU response;
  if (connection == NULL)
    response.BuildBandwidthReject(brq.m_requestSeqNum,
                                  H225_BandRejectReason::e_invalidConferenceID);
  else {
    if (connection->SetBandwidthAvailable(brq.m_bandWidth))
      response.BuildBandwidthConfirm(brq.m_requestSeqNum, brq.m_bandWidth);
    else
      response.BuildBandwidthReject(brq.m_requestSeqNum,
                                    H225_BandRejectReason::e_insufficientResources);
    connection->Unlock();
  }

  return response.Write(*transport);
}


BOOL H323Gatekeeper::InfoRequestResponse(const H323Connection * connection,
                                         unsigned seqNum)
{
  H323RasPDU response;
  H225_InfoRequestResponse & irr = response.BuildInfoRequestResponse(seqNum);

  endpoint.SetEndpointTypeInfo(irr.m_endpointType);
  irr.m_endpointIdentifier = endpointIdentifier;
  transport->SetUpTransportPDU(irr.m_rasAddress, TRUE);
  const H323ListenerList & listeners = endpoint.GetListeners();
  for (PINDEX i = 0; i < listeners.GetSize(); i++)
    listeners[i].SetUpTransportPDU(irr.m_callSignalAddress);

  if (endpoint.IsTerminal()) {
    irr.IncludeOptionalField(H225_InfoRequestResponse::e_endpointAlias);
    irr.m_endpointAlias.SetSize(1);
    H323SetAliasAddress(endpoint.GetLocalUserName(), irr.m_endpointAlias[0]);
  }

  if (connection != NULL) {
    irr.m_perCallInfo.SetSize(1);
    H225_InfoRequestResponse_perCallInfo_subtype & info = irr.m_perCallInfo[0];
    info.m_callReferenceValue = connection->GetCallReference();
    info.m_callIdentifier.m_guid = connection->GetCallIdentifier();
    info.m_conferenceID = connection->GetConferenceIdentifier();
    RTP_Session * session = connection->GetSession(RTP_Session::DefaultAudioSessionID);
    if (session != NULL) {
      info.IncludeOptionalField(H225_InfoRequestResponse_perCallInfo_subtype::e_audio);
      info.m_audio.SetSize(1);
      session->OnSendRasInfo(info.m_audio[0]);
    }
    session = connection->GetSession(RTP_Session::DefaultVideoSessionID);
    if (session != NULL) {
      info.IncludeOptionalField(H225_InfoRequestResponse_perCallInfo_subtype::e_video);
      info.m_video.SetSize(1);
      session->OnSendRasInfo(info.m_video[0]);
    }
    const H323Transport & controlChannel = connection->GetControlChannel();
    controlChannel.SetUpTransportPDU(info.m_h245.m_recvAddress, TRUE);
    controlChannel.SetUpTransportPDU(info.m_h245.m_sendAddress, FALSE);
    info.m_callType.SetTag(H225_CallType::e_pointToPoint);
    info.m_bandWidth = connection->GetBandwidthUsed();
    info.m_callModel.SetTag(H225_CallModel::e_direct);
  }

  return response.Write(*transport);
}


BOOL H323Gatekeeper::OnReceiveInfoRequest(const H225_InfoRequest & irq)
{
  H323GloballyUniqueID id = irq.m_callIdentifier.m_guid;
  H323Connection * connection = endpoint.FindConnectionWithLock(id.AsString());

  InfoRequestResponse(connection, irq.m_requestSeqNum);
  connection->Unlock();

  return TRUE;
}


#define DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
  void H323Gatekeeper::OnSend##name(H225_##name &) { }

#define DEFINE_GATEKEEPER_PROTOCOL_REQUEST(name) \
  DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
  BOOL H323Gatekeeper::OnReceive##name(const H225_##name &) { return FALSE; }

#define DEFINE_GATEKEEPER_PROTOCOL_REJECT(name, req) \
  DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
  BOOL H323Gatekeeper::OnReceive##name(const H225_##name & pdu) { \
    if (!CheckForResponse(H225_RasMessage::e_##req, pdu.m_requestSeqNum)) return FALSE; \
    PTRACE(1, "RAS\t" #name " rejected: " << pdu.m_rejectReason.GetTagName()); \
    responseResult = RejectReceived; return TRUE; \
  }

#define DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(name, req) \
  DEFINE_GATEKEEPER_PROTOCOL_SEND(name) \
  BOOL H323Gatekeeper::OnReceive##name(const H225_##name & pdu) { \
    return CheckForResponse(H225_RasMessage::e_##req, pdu.m_requestSeqNum); \
  }


DEFINE_GATEKEEPER_PROTOCOL_REQUEST(GatekeeperRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(GatekeeperConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(GatekeeperReject, gatekeeperRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(RegistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RegistrationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(RegistrationReject, registrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(UnregistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(UnregistrationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(UnregistrationReject, unregistrationRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(AdmissionRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(AdmissionConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(AdmissionReject, admissionRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(BandwidthRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(BandwidthConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(BandwidthReject, bandwidthRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(DisengageRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(DisengageConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(DisengageReject, disengageRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(LocationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(LocationConfirm);
DEFINE_GATEKEEPER_PROTOCOL_REJECT(LocationReject, locationRequest);
DEFINE_GATEKEEPER_PROTOCOL_SEND(InfoRequest);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestResponse, infoRequest);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(NonStandardMessage);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(UnknownMessageResponse);
DEFINE_GATEKEEPER_PROTOCOL_SEND(RequestInProgress);
DEFINE_GATEKEEPER_PROTOCOL_REQUEST(ResourcesAvailableIndicate);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(ResourcesAvailableConfirm, resourcesAvailableIndicate);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestAck, infoRequest);
DEFINE_GATEKEEPER_PROTOCOL_CONFIRM(InfoRequestNak, infoRequest);


void H323Gatekeeper::OnReceiveUnkown(const H323RasPDU &)
{
  H323RasPDU response;
  response.BuildUnknownMessageResponse(0);
  response.Write(*transport);
}

    
/////////////////////////////////////////////////////////////////////////////
