/*
 * rtp.cxx
 *
 * RTP protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: rtp.cxx,v $
 * Revision 1.8  1999/09/03 02:17:50  robertj
 * Added more debugging
 *
 * Revision 1.7  1999/08/31 12:34:19  robertj
 * Added gatekeeper support.
 *
 * Revision 1.6  1999/07/16 02:13:54  robertj
 * Slowed down the control channel statistics packets.
 *
 * Revision 1.5  1999/07/13 09:53:24  robertj
 * Fixed some problems with jitter buffer and added more debugging.
 *
 * Revision 1.4  1999/07/09 06:09:52  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.3  1999/06/22 13:49:40  robertj
 * Added GSM support and further RTP protocol enhancements.
 *
 * Revision 1.2  1999/06/14 08:42:52  robertj
 * Fixed bug in dropped packets display was negative.
 *
 * Revision 1.1  1999/06/14 06:12:25  robertj
 * Changes for using RTP sessions correctly in H323 Logical Channel context
 *
 */

#include <ptlib.h>
#include "rtp.h"
#include "h323.h"
#include "h225.h"
#include "h245.h"


#define new PNEW

#define UDP_BUFFER_SIZE 32768



/////////////////////////////////////////////////////////////////////////////

RTP_DataFrame::RTP_DataFrame(PINDEX sz)
  : PBYTEArray(12+sz)
{
  payloadSize = sz;
  theArray[0] = '\x80';
}


void RTP_DataFrame::SetExtension(BOOL ext)
{
  if (ext)
    theArray[0] |= 0x10;
  else
    theArray[0] &= 0xef;
}


void RTP_DataFrame::SetMarker(BOOL m)
{
  if (m)
    theArray[1] |= 0x80;
  else
    theArray[1] &= 0x7f;
}


void RTP_DataFrame::SetPayloadType(unsigned t)
{
  PAssert(t <= 0x7f, PInvalidParameter);

  theArray[1] &= 0x80;
  theArray[1] |= t;
}


DWORD RTP_DataFrame::GetContribSource(PINDEX idx) const
{
  PAssert(idx < GetContribSrcCount(), PInvalidParameter);
  return ((PUInt32b *)&theArray[12])[idx];
}


void RTP_DataFrame::SetContribSource(PINDEX idx, DWORD src)
{
  PAssert(idx <= 15, PInvalidParameter);

  if (idx >= GetContribSrcCount()) {
    BYTE * oldPayload = GetPayloadPtr();
    theArray[0] &= 0xf0;
    theArray[1] |= idx+1;
    SetSize(GetHeaderSize()+payloadSize);
    memmove(GetPayloadPtr(), oldPayload, payloadSize);
  }

  ((PUInt32b *)&theArray[12])[idx] = src;
}


BOOL RTP_DataFrame::SetPayloadSize(PINDEX sz)
{
  payloadSize = sz;
  return SetMinSize(GetHeaderSize()+payloadSize);
}


/////////////////////////////////////////////////////////////////////////////

RTP_ControlFrame::RTP_ControlFrame(PINDEX sz)
{
  SetPayloadSize(sz);
  theArray[0] = '\x80';
}


void RTP_ControlFrame::SetPayloadType(unsigned t)
{
  PAssert(t < 256, PInvalidParameter);
  theArray[1] = (BYTE)t;
}


void RTP_ControlFrame::SetPayloadSize(PINDEX sz)
{
  sz = (sz+3)/4;
  PAssert(sz <= 0xffff, PInvalidParameter);

  SetMinSize(4*sz+1);
  *(PUInt16b *)&theArray[2] = (WORD)sz;
}


/////////////////////////////////////////////////////////////////////////////

RTP_Session::RTP_Session(const H323Connection & conn, unsigned id)
  : connection(conn)
{
  PAssert(id > 0 && id < 256, PInvalidParameter);
  sessionID = (BYTE)id;

  referenceCount = 1;

  expectedSequenceNumber = 0;
  lastSequenceNumber = (WORD)rand();

  syncSourceOut = rand()*rand();
  senderReportInterval = 250;  // Number of data packets between reports

  packetsSent = 0;
  octetsSent = 0;
  packetsReceived = 0;
  packetsLost = 0;
  jitterLevel = 0;
}


void RTP_Session::OnSendRasInfo(H225_RTPSession & info)
{
  info.m_sessionId = sessionID;
  info.m_ssrc = syncSourceOut;
}


#ifdef PTRACING
static void TraceReceiverReport(const char * prefixMessage,
                                const RTP_ControlFrame::ReceptionReport * reports,
                                PINDEX count)
{
  for (PINDEX i = 0; i < count; i++)
    PTRACE(3, "RTP\t" << prefixMessage << "ReceiverReport " << i << ":"
              " ssrc=" << reports[i].ssrc
           << " fraction=" << (unsigned)reports[i].fraction
           << " lost=" << reports[i].GetLostPackets()
           << " last_seq=" << reports[i].last_seq
           << " jitter=" << reports[i].jitter
           << " lsr=" << reports[i].lsr
           << " dlsr=" << reports[i].dlsr);
}


static void TraceSenderReport(const char * prefixMessage,
                              const RTP_ControlFrame::SenderReport & sender,
                              const RTP_ControlFrame::ReceptionReport * reports,
                              PINDEX count)
{
  PTRACE(3, "RTP\t" << prefixMessage << "SenderReport: "
              " ssrc=" << sender.ssrc
           << " ntp=" << sender.ntp_sec
           << " rtp=" << sender.rtp_ts
           << " psent=" << sender.psent
           << " osent=" << sender.osent
           << " rr_cnt=" << count);
  TraceReceiverReport(prefixMessage, reports, count);
}
#endif


BOOL RTP_Session::OnSendData(RTP_DataFrame & frame)
{
  frame.SetSequenceNumber(++lastSequenceNumber);
  frame.SetSyncSource(syncSourceOut);

  packetsSent++;
  if (packetsSent%senderReportInterval == 0) {
    RTP_ControlFrame report;
    report.SetPayloadType(RTP_ControlFrame::e_SenderReport);
    report.SetPayloadSize(sizeof(RTP_ControlFrame::SenderReport));

    RTP_ControlFrame::SenderReport * sender =
                              (RTP_ControlFrame::SenderReport *)report.GetPayloadPtr();
    sender->ssrc = syncSourceOut;
    sender->ntp_sec = ::time(NULL)+(70*365+8)*24*60*60U;
    sender->ntp_frac = rand();
    sender->rtp_ts = lastSequenceNumber;
    sender->psent = packetsSent;
    sender->osent = octetsSent;

    if (!WriteControl(report))
      return FALSE;

#ifdef PTRACING
    TraceSenderReport("Sent", *sender, NULL, 0);
#endif
  }

  return TRUE;
}


BOOL RTP_Session::OnReceiveData(const RTP_DataFrame & frame)
{
  // Check that the PDU is the right version
  if (frame.GetVersion() != RTP_DataFrame::ProtocolVersion)
    return FALSE;

  // Check for if a control packet rather than data packet.
  if (frame.GetPayloadType() > RTP_DataFrame::MaxPayloadType)
    return FALSE;

  // Check packet sequence numbers
  if (expectedSequenceNumber == 0) {
    expectedSequenceNumber = (WORD)(frame.GetSequenceNumber() + 1);
    PTRACE(3, "RTP\tFirst data:"
              " ver=" << frame.GetVersion()
           << " pt=" << frame.GetPayloadType()
           << " psz=" << frame.GetPayloadSize()
           << " m=" << frame.GetMarker()
           << " x=" << frame.GetExtension()
           << " seq=" << frame.GetSequenceNumber()
           << " ts=" << frame.GetTimestamp()
           << " src=" << frame.GetSyncSource()
           << " ccnt=" << frame.GetContribSrcCount());
  }
  else {
    WORD sequenceNumber = frame.GetSequenceNumber();
    if (sequenceNumber == expectedSequenceNumber)
      expectedSequenceNumber++;
    else if (sequenceNumber < expectedSequenceNumber) {
      PTRACE(3, "RTP\tOut of order packet received: "
             << sequenceNumber << " < " << expectedSequenceNumber);
      // This implementation simply ignores out of order packets
      return FALSE;
    }
    else {
      unsigned dropped = sequenceNumber - expectedSequenceNumber;
      packetsLost += dropped;
      PTRACE(3, "RTP\t" << dropped << " packets dropped.");
      expectedSequenceNumber = (WORD)(sequenceNumber + 1);
    }
  }

  packetsReceived++;

  return TRUE;
}


BOOL RTP_Session::OnReceiveControl(const RTP_ControlFrame & frame)
{
  BYTE * payload = frame.GetPayloadPtr();
  switch (frame.GetPayloadType()) {
    case RTP_ControlFrame::e_SenderReport :
      OnRxSenderReport(*(const RTP_ControlFrame::SenderReport *)payload,
                       (const RTP_ControlFrame::ReceptionReport *)(payload+sizeof(RTP_ControlFrame::SenderReport)),
                       frame.GetCount());
      break;

    case RTP_ControlFrame::e_ReceiverReport :
      if (frame.GetPayloadSize() < 4)
        PTRACE(2, "RTP\tReceiverReport packet truncated");
      else
        OnRxReceiverReport(*(const PUInt32b *)payload,
                           (const RTP_ControlFrame::ReceptionReport *)(payload+sizeof(PUInt32b)),
                           frame.GetCount());
      break;

    case RTP_ControlFrame::e_SourceDescription :
      OnRxSourceDescription(*(const RTP_ControlFrame::SourceDescription *)payload);
      break;

    case RTP_ControlFrame::e_Goodbye :
    {
      PString str;
      PINDEX count = frame.GetCount()*4;
      if (frame.GetPayloadSize() > count)
        str = PString((const char *)(payload+count+4), *payload);
      OnRxGoodbye((const PUInt32b *)payload, count, str);
      break;
    }

    case RTP_ControlFrame::e_ApplDefined :
    {
      PString str((const char *)(payload+4), 4);
      OnRxApplDefined(str, frame.GetCount(), *(const PUInt32b *)payload,
                      payload+8, frame.GetPayloadSize()-8);
      break;
    }

    default :
      PTRACE(2, "RTP\tUnknown control payload type: " << frame.GetPayloadType());
      return FALSE;
  }

  return TRUE;
}


void RTP_Session::OnRxSenderReport(const RTP_ControlFrame::SenderReport & sender,
                                   const RTP_ControlFrame::ReceptionReport * reports,
                                   PINDEX count)
{
#ifdef PTRACING
  TraceSenderReport("OnRx", sender, reports, count);
#endif
}


void RTP_Session::OnRxReceiverReport(DWORD src,
                                     const RTP_ControlFrame::ReceptionReport * reports,
                                     PINDEX count)
{
  PTRACE(3, "RTP\tOnReceiverReport: ssrc=" << src << " rr_cnt=" << count);
#ifdef PTRACING
  TraceReceiverReport("OnRx", reports, count);
#endif
}


void RTP_Session::OnRxSourceDescription(const RTP_ControlFrame::SourceDescription & description)
{
  PTRACE(3, "RTP\tOnSourceDescription: " << description.src);
}


void RTP_Session::OnRxGoodbye(const PUInt32b * /*ssrc*/, PINDEX count,
                              const PString & reason)
{
  PTRACE(3, "RTP\tOnGoodbye: \"" << reason << "\" src_cnt=" << count);
}


void RTP_Session::OnRxApplDefined(const PString & type, unsigned subtype, DWORD src,
                                  const BYTE * /*data*/, PINDEX size)
{
  PTRACE(3, "RTP\tOnApplDefined: \"" << type << "\"-" << subtype
                          << " " << src << " [" << size << ']');
}


/////////////////////////////////////////////////////////////////////////////

static void SetMinBufferSize(PUDPSocket & sock, int buftype)
{
  int sz = 0;
  if (sock.GetOption(buftype, sz))
    if (sz >= UDP_BUFFER_SIZE)
      return;

  if (!sock.SetOption(buftype, UDP_BUFFER_SIZE))
    PTRACE(1, "RTP\tSetOption(" << buftype << ") failed: " << sock.GetErrorText());
}


RTP_UDP::RTP_UDP(const H323Connection & conn, unsigned id, WORD portBase)
  : RTP_Session(conn, id)
{
  WORD nextPort = (WORD)(portBase&0xfffe);
  while (!dataSocket.Listen(1, nextPort) || !controlSocket.Listen(1, (WORD)(nextPort+1))) {
    dataSocket.Close();
    controlSocket.Close();
    nextPort += 2;
  }

#ifndef __BEOS__
  SetMinBufferSize(dataSocket,    SO_RCVBUF);
  SetMinBufferSize(dataSocket,    SO_SNDBUF);
  SetMinBufferSize(controlSocket, SO_RCVBUF);
  SetMinBufferSize(controlSocket, SO_SNDBUF);
#endif

  PTRACE(2, "RTP_UDP\tSession " << id
         << " created: data=" << dataSocket.GetPort()
         << " control=" << controlSocket.GetPort()
         << " ssrc=" << syncSourceOut);
}


RTP_UDP::~RTP_UDP()
{
  Close();
}


void RTP_UDP::Close()
{
  PTRACE(3, "RTP_UDP\tClosing sockets "
         << dataSocket.GetHandle() << ' ' << controlSocket.GetHandle());
  dataSocket.Close();
  controlSocket.Close();
}


BOOL RTP_UDP::OnSendingPDU(H245_H2250LogicalChannelParameters & param,
                           BOOL includeMediaChannel) const
{
  PTRACE(3, "RTP\tOnSendingPDU");

  param.m_sessionID = GetSessionID();

  param.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaGuaranteedDelivery);
  param.m_mediaGuaranteedDelivery = FALSE;

  param.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_silenceSuppression);
  param.m_silenceSuppression = FALSE;

  const H323Transport & transport = connection.GetControlChannel();

  // unicast must have mediaControlChannel
  param.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel);
  transport.SetUpTransportPDU(param.m_mediaControlChannel, controlSocket.GetPort());

  if (includeMediaChannel) {
    // set mediaChannel
    param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);
    transport.SetUpTransportPDU(param.m_mediaChannel, dataSocket.GetPort());
  }

  return TRUE;
}


void RTP_UDP::OnSendingAckPDU(H245_H2250LogicalChannelAckParameters & param) const
{
  PTRACE(3, "RTP\tOnSendingAckPDU");

  const H323Transport & transport = connection.GetControlChannel();

  // set mediaControlChannel
  param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel);
  transport.SetUpTransportPDU(param.m_mediaControlChannel, controlSocket.GetPort());

  // set mediaChannel
  param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);
  transport.SetUpTransportPDU(param.m_mediaChannel, dataSocket.GetPort());
}


static BOOL ExtractTransport(const H245_TransportAddress & pdu,
                             PIPSocket::Address & ipnum, WORD & port,
                             unsigned & errorCode)
{
  if (pdu.GetTag() != H245_TransportAddress::e_unicastAddress) {
    PTRACE(1, "RTP_UDP\tOnly unicast supported at this time");
    errorCode = H245_OpenLogicalChannelReject_cause::e_multicastChannelNotAllowed;
    return FALSE;
  }

  const H245_UnicastAddress & unicast = pdu;
  if (unicast.GetTag() != H245_UnicastAddress::e_iPAddress) {
    PTRACE(1, "RTP_UDP\tLogic error, UDP must be IP unicast");
    errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
    return FALSE;
  }

  const H245_UnicastAddress_iPAddress & addr = unicast;

  ipnum = PIPSocket::Address(addr.m_network[0],
                             addr.m_network[1],
                             addr.m_network[2],
                             addr.m_network[3]);

  port = (WORD)(unsigned)addr.m_tsapIdentifier;
  return TRUE;
}


BOOL RTP_UDP::OnReceivedPDU(const H245_H2250LogicalChannelParameters & param,
                            unsigned & errorCode)
{
  if (param.m_sessionID != GetSessionID()) {
    PTRACE(1, "RTP_UDP\tOpen for invalid session: " << param.m_sessionID);
    errorCode = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
    return FALSE;
  }

  if (!param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel)) {
    PTRACE(1, "RTP_UDP\tNo mediaControlChannel specified");
    errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
    return FALSE;
  }

  if (!ExtractTransport(param.m_mediaControlChannel, remoteAddress, remoteControlPort, errorCode))
    return FALSE;

  if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel))
    return ExtractTransport(param.m_mediaChannel, remoteAddress, remoteDataPort, errorCode);

  remoteDataPort = (WORD)(remoteControlPort - 1);
  return TRUE;
}


BOOL RTP_UDP::OnReceivedAckPDU(const H245_H2250LogicalChannelAckParameters & param)
{
  if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_sessionID)) {
    PTRACE(1, "RTP_UDP\tNo session specified");
    return FALSE;
  }

  if (param.m_sessionID != GetSessionID()) {
    PTRACE(1, "RTP_UDP\tAck for invalid session: " << param.m_sessionID);
    return FALSE;
  }

  if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel)) {
    PTRACE(1, "RTP_UDP\tNo mediaControlChannel specified");
    return FALSE;
  }

  unsigned errorCode;
  if (!ExtractTransport(param.m_mediaControlChannel, remoteAddress, remoteControlPort, errorCode))
    return FALSE;

  if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel)) {
    PTRACE(1, "RTP_UDP\tNo mediaChannel specified");
    return FALSE;
  }

  if (!ExtractTransport(param.m_mediaChannel, remoteAddress, remoteDataPort, errorCode))
    return FALSE;

  return TRUE;
}


void RTP_UDP::OnSendRasInfo(H225_RTPSession & info)
{
  RTP_Session::OnSendRasInfo(info);

  const H323Transport & transport = connection.GetControlChannel();

  transport.SetUpTransportPDU(info.m_rtpAddress.m_recvAddress, dataSocket.GetPort());
  H323TransportAddress(remoteAddress, remoteDataPort).SetPDU(info.m_rtpAddress.m_sendAddress);

  transport.SetUpTransportPDU(info.m_rtcpAddress.m_recvAddress, controlSocket.GetPort());
  H323TransportAddress(remoteAddress, remoteDataPort).SetPDU(info.m_rtcpAddress.m_sendAddress);
}


BOOL RTP_UDP::ReadData(RTP_DataFrame & frame)
{
  BOOL ok = TRUE;

  while (ok) {
    int selectStatus = PSocket::Select(dataSocket, controlSocket);
    switch (selectStatus) {
      case -1 :
        return ReadDataPDU(frame);

      case -2 :
        ok = ReadControlPDU();
        break;

      case -3 :
        ReadControlPDU();
        return ReadDataPDU(frame);

      case PSocket::Interrupted:
        PTRACE(3, "RTP\tInterrupted");
        ok = FALSE;
        break;

      default :
        PTRACE(1, "RTP\tSelect error: "
                << PChannel::GetErrorText((PChannel::Errors)selectStatus));
        ok = FALSE;
    }
  }

  return FALSE;
}


BOOL RTP_UDP::ReadDataPDU(RTP_DataFrame & frame)
{
  PIPSocket::Address addr;
  WORD port;

  while (dataSocket.ReadFrom(frame.GetPointer(), frame.GetSize(), addr, port)) {
    // Check recieved PDU is big enough
    PINDEX pduSize = dataSocket.GetLastReadCount();
    if (pduSize < RTP_DataFrame::MinHeaderSize || pduSize < frame.GetHeaderSize())
      PTRACE(2, "RTP\tReceived data packet too small: " << pduSize << " bytes");
    else {
      frame.SetPayloadSize(pduSize - frame.GetHeaderSize());
      if (OnReceiveData(frame))
        return TRUE;
    }
  }

  PTRACE(1, "RTP\tData read error: " << dataSocket.GetErrorText());
  dataSocket.Close();
  return FALSE;
}


BOOL RTP_UDP::ReadControlPDU()
{
  RTP_ControlFrame frame(2048);
  PIPSocket::Address addr;
  WORD port;

  while (controlSocket.ReadFrom(frame.GetPointer(), frame.GetSize(), addr, port)) {
    PINDEX pduSize = controlSocket.GetLastReadCount();
    if (pduSize < 4 || pduSize < 4+frame.GetPayloadSize())
      PTRACE(2, "RTP\tReceived control packet too small: " << pduSize << " bytes");
    else {
      if (OnReceiveControl(frame))
        return TRUE;
    }
  }

  PTRACE(1, "RTP\tControl read error: " << controlSocket.GetErrorText());
  controlSocket.Close();
  return FALSE;
}


BOOL RTP_UDP::WriteData(RTP_DataFrame & frame)
{
  if (!OnSendData(frame))
    return FALSE;

  if (dataSocket.WriteTo(frame.GetPointer(),
                         frame.GetHeaderSize()+frame.GetPayloadSize(),
                         remoteAddress, remoteDataPort))
    return TRUE;

  PTRACE(1, "RTP\tWrite error on data port: " << dataSocket.GetErrorText());
  return FALSE;
}


BOOL RTP_UDP::WriteControl(RTP_ControlFrame & frame)
{
  if (controlSocket.WriteTo(frame.GetPointer(), frame.GetPayloadSize()+4,
                            remoteAddress, remoteControlPort))
    return TRUE;

  PTRACE(1, "RTP\tWrite error on control port: " << controlSocket.GetErrorText());
  return FALSE;
}


/////////////////////////////////////////////////////////////////////////////
